#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Thu Dec  3 14:42:29 2020
#
#@author: Majdi
#"""

import random
import numpy as np
from numpy import arange, multiply, zeros, copy
import joblib
from neorl.hybrid.edevcore.helpers import ARCHIVE, generator, sub2ind, gnR1R2, randFCR
from neorl.evolu.discrete import mutate_discrete, encode_grid_to_discrete 
from neorl.evolu.discrete import decode_discrete_to_grid, encode_grid_indv_to_discrete
from neorl.utils.seeding import set_neorl_seed
from neorl.utils.tools import get_population, check_mixed_individual


class EDEV(object):
    """
    Ensemble of differential evolution variants
    
    :param mode: (str) problem type, either ``min`` for minimization problem or ``max`` for maximization
    :param bounds: (dict) input parameter type and lower/upper bounds in dictionary form. Example: ``bounds={'x1': ['int', 1, 4], 'x2': ['float', 0.1, 0.8], 'x3': ['float', 2.2, 6.2]}``
    :param fit: (function) the fitness function 
    :param npop: (int): total size of the full population, which will be divided into three sub-populations
    :param int_transform: (str): method of handling int/discrete variables, choose from: ``nearest_int``, ``sigmoid``, ``minmax``.
    :param ncores: (int) number of parallel processors (must be ``<= npop``)
    :param seed: (int) random seed for sampling
    """
    #:param lambda_: (float): fraction of ``npop`` to split into 3 sub-populations. ``pop1 = npop - npop * lambda_``, ``pop2 = pop3 =  npop * lambda_`` (see **Notes** below)        
    def __init__(self, mode, bounds, fit, npop=100, 
                 int_transform='nearest_int', ncores=1, seed=None):

        set_neorl_seed(seed)
        
        lambda_=0.1   #for EDEV stability, better to keep this parameter fixed. 
        assert npop >= 70, '--error: EDEV ensemble requires a large population of 70 or more'
        assert ncores <= npop, '--error: ncores ({}) must be less than or equal than npop ({})'.format(ncores, npop)
        assert 0 < lambda_ < 1, '--error: lambda_ must be more than 0 and less than 1'
        #--mir
        self.mode=mode
        if mode == 'min':
            self.fit=fit
        elif mode == 'max':
            def fitness_wrapper(*args, **kwargs):
                return -fit(*args, **kwargs) 
            self.fit=fitness_wrapper
        else:
            raise ValueError('--error: The mode entered by user is invalid, use either `min` or `max`')
          
        self.int_transform=int_transform
        if int_transform not in ["nearest_int", "sigmoid", "minmax"]:
            raise ValueError('--error: int_transform entered by user is invalid, must be `nearest_int`, `sigmoid`, or `minmax`')
            
        self.bounds=bounds
        self.ncores = ncores
        self.npop=npop
        self.lambda_=lambda_
        
        #infer variable types 
        self.var_type = np.array([bounds[item][0] for item in bounds])
        
        #mir-grid
        if "grid" in self.var_type:
            self.grid_flag=True
            self.orig_bounds=bounds  #keep original bounds for decoding
            #print('--debug: grid parameter type is found in the space')
            self.bounds, self.bounds_map=encode_grid_to_discrete(self.bounds) #encoding grid to int
            #define var_types again by converting grid to int
            self.var_type = np.array([self.bounds[item][0] for item in self.bounds])
        else:
            self.grid_flag=False
            self.bounds = bounds
            self.orig_bounds=bounds
        
        self.dim = len(bounds)
        self.lb=np.array([self.bounds[item][1] for item in self.bounds])
        self.ub=np.array([self.bounds[item][2] for item in self.bounds])
        
        slice_index=int(self.lambda_*self.npop)
        if slice_index <= 6:
            raise Exception('--error: the size of pop2 and pop3 ({}) must be more than 6 individuals. Make sure int(lambda_ * npop) >= 7'.format(slice_index))

    def init_sample(self, bounds):
    
        indv=[]
        for key in bounds:
            if bounds[key][0] == 'int':
                indv.append(random.randint(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'float':
                indv.append(random.uniform(bounds[key][1], bounds[key][2]))
            #elif bounds[key][0] == 'grid':
            #    indv.append(random.sample(bounds[key][1],1)[0])
            else:
                raise Exception ('unknown data type is given, either int, float, or grid are allowed for parameter bounds')   
        return np.array(indv)

    def eval_pop(self, pop_to_evaluate):
    
        #---------------------
        # Fitness calcs
        #---------------------
        core_lst=[]
        for case in range (0, pop_to_evaluate.shape[0]):
            core_lst.append(pop_to_evaluate[case, :])
    
        if self.ncores > 1:

            with joblib.Parallel(n_jobs=self.ncores) as parallel:
                fitness_lst=parallel(joblib.delayed(self.fit_worker)(item) for item in core_lst)
                
        else:
            fitness_lst=[]
            for item in core_lst:
                fitness_lst.append(self.fit_worker(item))
        
        return fitness_lst

    def select(self, pos, fit):
        
        best_fit=np.min(fit)
        min_idx=np.argmin(fit)
        best_pos=pos[min_idx,:].copy()
        
        return best_pos, best_fit , min_idx
        
    def ensure_bounds(self, vec):
    
        vec_new = []
        # cycle through each variable in vector 
        for i, (key, val) in enumerate(self.bounds.items()):
    
            # variable exceedes the minimum boundary
            if vec[i] < self.bounds[key][1]:
                vec_new.append(self.bounds[key][1])
    
            # variable exceedes the maximum boundary
            if vec[i] > self.bounds[key][2]:
                vec_new.append(self.bounds[key][2])
    
            # the variable is fine
            if self.bounds[key][1] <= vec[i] <= self.bounds[key][2]:
                vec_new.append(vec[i])
            
        return vec_new
    
    def fit_worker(self, x):
        #This worker is for parallel calculations
        
        # Clip the position outside the lower/upper bounds and return same position
        x=self.ensure_bounds(x)

        
        if self.grid_flag:
            #decode the individual back to the int/float/grid mixed space
            x=decode_discrete_to_grid(x,self.orig_bounds,self.bounds_map)
        
        # Calculate objective function for each search agent
        fitness = self.fit(x)
        self.FES += 1
        
        return fitness
    
    def ensure_discrete(self, vec):
        #"""
        #to mutate a vector if discrete variables exist 
        #handy function to be used within SSA phases

        #Params:
        #vec - position in vector/list form

        #Return:
        #vec - updated position vector with discrete values
        #"""
        
        for dim in range(self.dim):
            if self.var_type[dim] == 'int':
                vec[dim] = mutate_discrete(x_ij=vec[dim], 
                                               x_min=min(vec),
                                               x_max=max(vec),
                                               lb=self.lb[dim], 
                                               ub=self.ub[dim],
                                               alpha=self.a,
                                               method=self.int_transform,
                                               )
        
        return vec
            

    def evolute(self, ngen, ng=20, x0=None, verbose=False):
        """
        This function evolutes the EDEV algorithm for a number of generations.
        
        :param ngen: (int) number of generations to evolute
        :param ng: (int) the period or number of generations to determine the best performing DE variant and reward subpopulation assignment, ``ng < ngen`` (see **Notes** below for more info).
        :param x0: (list of lists) initial position of the individuals (must be of same size as ``npop``)
        :param verbose: (bool) print statistics to screen
        
        :return: (tuple) (best individual, best fitness, and dictionary containing major search results)
        """
        
        assert ng < ngen, '--error: ng must be less than ngen to allow frequent updates'
        self.history = {'global_fitness':[], 'JADE':[], 'CoDE':[], 'EPSDE':[]}
        self.best_fitness=float("inf") 
        self.verbose=verbose
        self.a=1   #for discrete analysis

        self.mixPop = np.zeros((self.npop, self.dim))
        if x0:
            assert len(x0) == self.npop, '--error: the length of x0 ({}) MUST equal the number of npop in the group ({})'.format(len(x0), self.npop)
                
            for i in range(self.npop):
                check_mixed_individual(x=x0[i], bounds=self.orig_bounds) #assert the type provided is consistent
                if self.grid_flag:
                    self.mixPop[i,:] = encode_grid_indv_to_discrete(x0[i], bounds=self.orig_bounds, bounds_map=self.bounds_map)
                else:
                    self.mixPop[i,:] = x0[i]
        else:
            # Initialize the positions
            for i in range(self.npop):
                self.mixPop[i,:]=self.init_sample(self.bounds)
        

        #---------------------------------------------------------------------------
        #-------------------------------EDEV starts here----------------------------
        #---------------------------------------------------------------------------
        # Define the dimension of the problem
        self.FES=0
        arrayGbestChange=np.ones((3), dtype=float)
        arrayGbestChangeRate=np.zeros((3), dtype=float)
        indexBestLN=0
        numViaLN=np.array([0,0,0])
        rateViaLN=zeros((ngen,len(numViaLN)))
        
        #evaluate the first population   
        #for discrete mutation
        for iii in range(self.mixPop.shape[0]):
            self.mixPop[iii, :] = self.ensure_bounds(self.mixPop[iii, :])
            self.mixPop[iii, :] = self.ensure_discrete(self.mixPop[iii, :])
        mixVal=np.array(self.eval_pop(self.mixPop))
        self.best_position, self.overallBestVal, _ = self.select(self.mixPop, mixVal)
        
        permutation=np.random.permutation(self.npop)
                
        slice_index=int(self.lambda_*self.npop)
        arrayThird=permutation[:slice_index] #for EPSDE
        arraySecond=permutation[slice_index : 2*slice_index]  #for CoDE
        arrayFirst=permutation[2*slice_index:]    #for JADE 
        
        if verbose:
            print('--debug: initial size of EDEV population is %i'%self.npop)
            print('--debug: initial size of pop1 is %i'%len(arrayFirst))
            print('--debug: initial size of pop2 is %i'%len(arraySecond))
            print('--debug: initial size of pop3 is %i'%len(arrayThird))
        
        ## Initialize JADE related
        popold=self.mixPop[arrayFirst,:]
        valParents=mixVal[arrayFirst]
        c=1 / 10
        pj=0.05
        CRm=0.5
        Fm=0.5
        
        #initialize archive
        archive=ARCHIVE(NP=len(arrayFirst), dim=self.dim)
        
        #the values and indices of the best solutions
        #valBest=np.sort(valParents)
        indBest=np.argsort(valParents)
        
        #Initialize CoDE related
        popCODE=self.mixPop[arraySecond,:]
        valCODE=mixVal[arraySecond]
        
        #Initialize EPSDE related
        I_D=self.dim
        FM_pop=self.mixPop[arrayThird,:]
        FM_popold=zeros(FM_pop.shape)
        I_NP=len(arrayThird)
        val=zeros((I_NP))
        self.FVr_bestmem=zeros((I_D))
        #FVr_bestmemit=zeros((I_D))
        I_nfeval=0
        RATE=0
        
        #------Evaluate the best member after initialization----------------------
        #for discrete mutation
        for iii in range(FM_pop.shape[0]):
            FM_pop[iii, :] = self.ensure_bounds(FM_pop[iii, :])
            FM_pop[iii, :] = self.ensure_discrete(FM_pop[iii, :])
            
        val=np.array(self.eval_pop(FM_pop))
        self.FVr_bestmem, F_bestval, I_best_index=self.select(pos=FM_pop, fit=val)
        
        #------DE-Minimization---------------------------------------------
        #------FM_popold is the population which has to compete. It is--------
        #------static through one iteration. FM_pop is the newly--------------
        #------emerging population.----------------------------------------
        I_iter=0
        FF=np.array([0.4, 0.5, 0.6, 0.7, 0.8, 0.9])
        CR=np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])
        spara=np.random.permutation(3)
        PPara=np.empty((0,3), dtype=float)
        inde=np.arange(0,I_NP)
        RR=zeros((self.npop))
        gen=0
        FESj=0
        goodCR=[]
        goodF=[]
        rate=[]
        
        while gen < ngen:
            
            self.a= 1 - gen * ((1) / ngen)  #mir: a decreases linearly between 1 to 0, for discrete mutation
        
            ## update the index for adjust different algorithms
            if gen % ng == 0:
                arrayGbestChangeRate[0]=arrayGbestChange[0] / len(arrayFirst)
                arrayGbestChangeRate[1]=arrayGbestChange[1] / (len(arraySecond)*3)
                arrayGbestChangeRate[2]=arrayGbestChange[2] / len(arrayThird)
                indexBestLN=np.argmax(arrayGbestChangeRate)
        
                arrayGbestChange=np.ones((3), dtype=float)
                arrayGbestChangeRate=np.zeros((3), dtype=float)
        
            permutation=np.random.permutation(self.npop)
            if indexBestLN == 0:
                arrayThird=permutation[:slice_index]
                arraySecond=permutation[slice_index : 2*slice_index] 
                arrayFirst=permutation[2*slice_index:]
                numViaLN[0]=numViaLN[0] + 1
            else:
                if indexBestLN == 1:
                    arrayThird=permutation[:slice_index]
                    arrayFirst=permutation[slice_index : 2*slice_index] 
                    arraySecond=permutation[2*slice_index:]
                    numViaLN[1]=numViaLN[1] + 1
                else:
                    if indexBestLN == 2:
                        arrayFirst=permutation[:slice_index]
                        arraySecond=permutation[slice_index : 2*slice_index] 
                        arrayThird=permutation[2*slice_index:]
                        numViaLN[2]=numViaLN[2] + 1
                        
            rateViaLN[gen,:]=numViaLN / sum(numViaLN)
            
            #----------------------------
            #JADE
            #----------------------------
            pop=self.mixPop[arrayFirst,:]
            valParents=mixVal[arrayFirst]
            popsize=len(arrayFirst)
            if FESj > 1 and len(goodCR) > 0 and sum(goodF) > 0:
                CRm = (1 - c)*CRm + c * np.mean(goodCR)
                Fm = (1 - c)*Fm + c * sum(goodF ** 2) / sum(goodF)
        
            # Generate CR according to a normal distribution with mean CRm, and std 0.1
            # Generate F according to a cauchy distribution with location parameter Fm, and scale parameter 0.1
            Fj,CRj=randFCR(popsize,CRm,0.1,Fm,0.1)
            r0=arange(0,popsize)
            if archive.pop.size == 0:
                popAll=copy(pop)
            else:
                popAll=np.concatenate((pop, archive.pop),axis=0)
            r1,r2=gnR1R2(popsize-1,popAll.shape[0]-1,r0)   #-1 for python indexing from 0
            indBest=np.argsort(valParents)
            pNP=max(round(pj*popsize),2)   #choose at least two best solutions
            randindex=np.ceil(np.random.uniform(size=(popsize))*pNP)  #select from [1, 2, 3, ..., pNP]
            randindex[randindex < 1]=1   #to avoid the problem that rand = 0 and thus ceil(rand) = 0
            randindex=randindex-1 #for python indexing
            assert np.min(randindex) >= 0, '--error: one of the indices in randindex is less than 0, the alg does not function properly'
            pbest=pop[indBest[randindex.astype(int)],:]  #randomly choose one of the top 100p% solutions
            
            # == == == == == == == == == == == == == == == Mutation == == == == == == == == == == == == ==
            Ftr=Fj.reshape(Fj.shape[0],1)   #transform Fj
            Ftr=np.tile(Ftr,(1,self.dim))   #convert Fj from npop dimensions to a npop x dim with fixed value in all dimensions
            vi=pop + multiply(Ftr,(pbest - pop + pop[r1,:] - popAll[r2,:]))
            
            #apply bound constraint check
            for iii in range(len(vi)):
                vi[iii,:]=self.ensure_bounds(vi[iii,:])
                
            # == == == == = Crossover == == == == =
            CRtr=CRj.reshape(CRj.shape[0],1)   #transform Crj
            CRtr=np.tile(CRtr,(1,self.dim))   #convert Fj from npop dimensions to a npop x dim with fixed value in all dimensions
            
            mask=np.random.uniform(size=(popsize,self.dim)) > CRtr  #mask is used to indicate which elements of ui comes from the parent
            rows=arange(popsize)
            cols=np.array(np.floor(np.random.uniform(size=(popsize)) * self.dim) + 1, dtype=int)
            jrand=np.array(sub2ind((popsize,self.dim),rows,cols),dtype=int)
            mask=mask.reshape(-1)
            mask[jrand]=False
            mask=mask.reshape(popsize,self.dim)
            ui=copy(vi)
            ui[mask]=pop[mask]
            
            #evaluate the offspring
            #for discrete mutation
            for iii in range(ui.shape[0]):
                ui[iii, :] = self.ensure_bounds(ui[iii, :])
                ui[iii, :] = self.ensure_discrete(ui[iii, :])
            valOffspring=np.array(self.eval_pop(ui))
            
            #== == == == == == == == == == == == == == == Selection == == == == == == == == == == == == ==
            # I == 1: the parent is better; I == 2: the offspring is better
            valParents=np.minimum(valParents, valOffspring)
            I= np.where(valOffspring <= valParents)[0] #indices where offspring is better than the parents
            popold=copy(pop)
            archive.UpdateArchive(popold[I,:],valParents[I])
            popold[I,:]=ui[I,:]
            goodCR=CRj[I]
            goodCR=Fj[I]
            if np.min(valParents) < self.overallBestVal:
                self.overallBestVal=np.min(valParents)
                min_best_index=np.argmin(valParents)
                self.best_position=popold[min_best_index,:].copy()
                
            self.jade_fitness=np.min(valParents)  #store best JADE fitness
            
            arrayGbestChange[0]=arrayGbestChange[0] + np.sum(mixVal[arrayFirst] - valParents)
            self.mixPop[arrayFirst,:]=popold
            mixVal[arrayFirst]=valParents
            
            #----------------------------
            #CoDE
            #----------------------------
            popCODE=self.mixPop[arraySecond,:]
            valCODE=mixVal[arraySecond]
            popsizeC=len(arraySecond)
            pTemp=copy(popCODE)
            fitTemp=copy(valCODE)
            uSet=zeros((3*popsizeC,self.dim))
            for i in range(popsizeC):
                F=np.array([1.0,1.0,0.8])
                CRC=np.array([0.1,0.9,0.2])
                #Uniformly and randomly select one of the control
                # parameter settings for each trial vector generation strategy
                paraIndex=np.floor(np.random.uniform(size=(3))*len(F))
                u=generator(popCODE, self.bounds, i, F, CRC, popsizeC, self.dim, paraIndex.astype(int))
                uSet[(i+1)*3 - 3 : 3*(i+1),:]=copy(u)
                
            # Evaluate the trial vectors
            for iii in range(uSet.shape[0]):   #discrete mutation
                uSet[iii, :] = self.ensure_bounds(uSet[iii, :])
                uSet[iii, :] = self.ensure_discrete(uSet[iii, :])
            fitSet=np.array(self.eval_pop(uSet))
            
            for i in range(popsizeC):
                #minVal=np.min(fitSet[(i+1)*3 - 3 : 3*(i+1)])
                minID=np.argmin(fitSet[(i+1)*3 - 3 : 3*(i+1)])
                bestInd=uSet[3*i + minID,:]
                bestIndFit=fitSet[3*i + minID]
                # Choose the better one between the trial vector and the target vector
                if valCODE[i] >= bestIndFit:
                    pTemp[i,:]=bestInd.copy()
                    fitTemp[i]=bestIndFit

            popCODE=copy(pTemp)
            valCODE=copy(fitTemp)
                    
            if np.min(fitTemp) < self.overallBestVal:
                self.overallBestVal=np.min(fitTemp)
                min_best_index=np.argmin(fitTemp)
                self.best_position=popCODE[min_best_index,:].copy()
            
            self.code_fitness=np.min(fitTemp)  #store best CoDE fitness

            arrayGbestChange[1]=arrayGbestChange[1] + np.sum(mixVal[arraySecond] - valCODE)
            self.mixPop[arraySecond,:]=popCODE
            mixVal[arraySecond]=valCODE
            
            #----------------------------
            #EPSDE
            #----------------------------
            FM_pop=self.mixPop[arrayThird,:]
            val=mixVal[arrayThird]
            I_NP=len(arrayThird)
            inde=arange(I_NP)
            Para=zeros((I_NP,3))
            I_iter=gen
            
            if (I_iter == 0 or len(Para) < I_NP):
                Para[:,0]=spara[np.random.randint(spara.shape[0],size=inde.shape[0])]
                Para[:,1]=CR[np.random.randint(CR.shape[0],size=inde.shape[0])]
                Para[:,2]=FF[np.random.randint(FF.shape[0],size=inde.shape[0])]
            else:
                for k in range(len(inde)):
                    if (np.random.uniform() <= RATE and len(PPara) > 0) or len(RR) < I_NP:
                        RR[k]=np.random.randint(PPara.shape[0])
                        Para[inde[k],:]=PPara[np.random.randint(PPara.shape[0]),:]
                    else:
                        RR[k]=0
                        Para[inde[k],:]=[
                                        spara[np.random.randint(spara.shape[0])],
                                        CR[np.random.randint(CR.shape[0])],
                                        FF[np.random.randint(FF.shape[0])]
                                        ]
        
            RRR=[]
            count=0
            FM_popold=copy(FM_pop)
            par=np.zeros((I_NP, I_D))
            FM_ui=par.copy()
            
            for i in range(I_NP):
                FM_mui=np.random.uniform(size=(I_D)) < Para[i,1]
                
                dd=np.where(FM_mui == True)[0]
                if len(dd) == 0:
                    ddd=int(np.ceil(np.random.uniform() * I_D) - 1)
                    FM_mui[ddd]=1
                    
                FM_mpo=FM_mui < 0.5
                FM_bm=copy(self.FVr_bestmem)
                
                par[i,:]= np.random.normal(Para[i,2],0.001,size=(I_D))
                
                if (Para[i,0] == 0):
                    #DE/best/2/bin
                    ind=np.random.permutation(I_NP)
                    FM_pm3=FM_popold[ind[0],:]
                    FM_pm4=FM_popold[ind[1],:]
                    FM_pm5=FM_popold[ind[2],:]
                    FM_pm6=FM_popold[ind[3],:]
                    FM_ui[i,:]=FM_bm + (FM_pm3 - FM_pm4 + FM_pm5 - FM_pm6) * par[i,:]
                    FM_ui[i,:]=multiply(FM_popold[i,:],FM_mpo) + multiply(FM_ui[i,:],FM_mui)
                if (Para[i,0] == 1):
                    #DE/rand/1/bin
                    ind=np.random.permutation(I_NP)
                    FM_pm7=FM_popold[ind[0],:]
                    FM_pm8=FM_popold[ind[1],:]
                    FM_pm9=FM_popold[ind[2],:]
                    FM_ui[i,:]=FM_pm7 + par[i,:] * (FM_pm8 - FM_pm9)
                    FM_ui[i,:]=multiply(FM_popold[i,:],FM_mpo) + multiply(FM_ui[i,:],FM_mui)
                if (Para[i,0] == 2):
                    # DE/current-to-rand/1/bin/
                    ind=np.random.permutation(I_NP)
                    FM_pm21=FM_popold[ind[0],:]
                    FM_pm22=FM_popold[ind[1],:]
                    FM_pm23=FM_popold[ind[2],:]
                    FM_ui[i,:]=FM_popold[i,:] + np.random.uniform(size=I_D) * (FM_pm21 - FM_popold[i,:]) + par[i,:] * (FM_pm22 - FM_pm23)
        
                if np.any(FM_ui[i,:] < self.lb) or np.any(FM_ui[i,:] > self.ub):
                    #reset an individual outside the bounds
                    FM_ui[i,:]=self.lb + multiply((self.ub - self.lb),np.random.uniform(size=I_D))
                        
            #evaluate the updated population
            for iii in range(FM_ui.shape[0]):   #discrete mutation
                FM_ui[iii, :] = self.ensure_bounds(FM_ui[iii, :])
                FM_ui[iii, :] = self.ensure_discrete(FM_ui[iii, :])
            tempval=np.array(self.eval_pop(FM_ui))
                    
        
            for i in range(I_NP):
                I_nfeval=I_nfeval + 1
                if (tempval[i] < val[i]):
                    FM_pop[i,:]=FM_ui[i,:]
                    val[i]=tempval[i]
                    PPara=np.append(PPara, Para[i,:].reshape(1,-1), axis=0)
                    if RR[i] != 0:
                        RRR.append(int(RR[i]))
                    if tempval[i] < F_bestval:
                        F_bestval=tempval[i]
                        self.FVr_bestmem=FM_ui[i,:]
                        I_best_index=i
                else:
                    count=count + 1
                    
            if np.min(val) < self.overallBestVal:
                self.overallBestVal=np.min(val)
                min_best_index=np.argmin(val)
                self.best_position=FM_pop[min_best_index,:].copy()
            
            self.epsde_fitness=np.min(val)  #store best EPSDE fitness
                
            arrayGbestChange[2]=arrayGbestChange[2] + np.sum(mixVal[arrayThird] - val)
            self.mixPop[arrayThird,:]=FM_pop
            mixVal[arrayThird]=val
            PPara = np.delete(PPara, RRR, axis=0)
            rate.append(count / I_NP)
            if I_iter > 10:
                RATE=np.mean(rate[(I_iter - 10) : I_iter])
            else:
                RATE=np.mean(rate)
            I_iter=I_iter + 1
            gen=gen + 1

        #---------------------------------------------------------------------------
        #-------------------------------EDEV ends here----------------------------
        #---------------------------------------------------------------------------
                
            #--mir
            if self.mode=='max':
                self.fitness_best_correct=-self.overallBestVal
                self.jade_fitness=-self.jade_fitness
                self.code_fitness=-self.code_fitness
                self.epsde_fitness=-self.epsde_fitness
                
            else:
                self.fitness_best_correct=self.overallBestVal

            self.last_pop=self.mixPop.copy()
            self.last_fit=np.array(mixVal).copy()            
            self.history['global_fitness'].append(self.fitness_best_correct)
            self.history['JADE'].append(self.jade_fitness)
            self.history['CoDE'].append(self.code_fitness)
            self.history['EPSDE'].append(self.epsde_fitness)
                        
            # Print statistics
            if self.verbose:
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
                print('EDEV step {}/{}, npop={}, Ncores={}'.format((gen)*self.npop, ngen*self.npop, self.npop, self.ncores))
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
                print('Best EDEV Fitness:', np.round(self.fitness_best_correct,6))
                if self.grid_flag:
                    self.ind_decoded = decode_discrete_to_grid(self.best_position, self.orig_bounds, self.bounds_map)
                    print('Best Position:', self.ind_decoded)
                else:
                    print('Best Position:', self.best_position)
                print('JADE Fitness:', self.jade_fitness)
                print('CoDE Fitness:', self.code_fitness)
                print('EPSDE Fitness:', self.epsde_fitness)
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')

        #mir-grid
        if self.grid_flag:
            self.ind_correct = decode_discrete_to_grid(self.best_position, self.orig_bounds, self.bounds_map)
        else:
            self.ind_correct = self.best_position                

        if self.mode=='max':
            self.last_fit=-self.last_fit

        #--mir return the last population for restart calculations
        if self.grid_flag:
            self.history['last_pop'] = get_population(self.last_pop, fits=self.last_fit, grid_flag=True, 
                                                     bounds=self.orig_bounds, bounds_map=self.bounds_map)
        else:
            self.history['last_pop'] = get_population(self.last_pop, fits=self.last_fit, grid_flag=False)
            
        self.history['F-Evals'] = self.FES   #function evaluations 
        if self.verbose:
            print('------------------------ EDEV Summary --------------------------')
            print('Best fitness (y) found:', self.fitness_best_correct)
            print('Best individual (x) found:', self.ind_correct)
            print('--------------------------------------------------------------')  
            
        return self.ind_correct, self.fitness_best_correct, self.history




from __future__ import print_function

import numpy as np


# ================================================================
# Misc
# ================================================================


def fmt_row(width, row, header=False):
    """
    fits a list of items to at least a certain length

    :param width: (int) the minimum width of the string
    :param row: ([Any]) a list of object you wish to get the string representation
    :param header: (bool) whether or not to return the string as a header
    :return: (str) the string representation of all the elements in 'row', of length >= 'width'
    """
    out = " | ".join(fmt_item(x, width) for x in row)
    if header:
        out = out + "\n" + "-" * len(out)
    return out


def fmt_item(item, min_width):
    """
    fits items to a given string length

    :param item: (Any) the item you wish to get the string representation
    :param min_width: (int) the minimum width of the string
    :return: (str) the string representation of 'x' of length >= 'l'
    """
    if isinstance(item, np.ndarray):
        assert item.ndim == 0
        item = item.item()
    if isinstance(item, (float, np.float32, np.float64)):
        value = abs(item)
        if (value < 1e-4 or value > 1e+4) and value > 0:
            rep = "%7.2e" % item
        else:
            rep = "%7.5f" % item
    else:
        rep = str(item)
    return " " * (min_width - len(rep)) + rep


COLOR_TO_NUM = dict(
    gray=30,
    red=31,
    green=32,
    yellow=33,
    blue=34,
    magenta=35,
    cyan=36,
    white=37,
    crimson=38
)


def colorize(string, color, bold=False, highlight=False):
    """
    Colorize, bold and/or highlight a string for terminal print

    :param string: (str) input string
    :param color: (str) the color, the lookup table is the dict at console_util.color2num
    :param bold: (bool) if the string should be bold or not
    :param highlight: (bool) if the string should be highlighted or not
    :return: (str) the stylized output string
    """
    attr = []
    num = COLOR_TO_NUM[color]
    if highlight:
        num += 10
    attr.append(str(num))
    if bold:
        attr.append('1')
    return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), string)



import re
from functools import reduce

import tensorflow as tf
import numpy as np

from neorl.rl.baselines.acktr.kfac_utils import detect_min_val, factor_reshape, gmatmul

KFAC_OPS = ['MatMul', 'Conv2D', 'BiasAdd']
KFAC_DEBUG = False


class KfacOptimizer:
    def __init__(self, learning_rate=0.01, momentum=0.9, clip_kl=0.01, kfac_update=2, stats_accum_iter=60,
                 full_stats_init=False, cold_iter=100, cold_lr=None, async_eigen_decomp=False,
                 async_stats=False, epsilon=1e-2, stats_decay=0.95, blockdiag_bias=False,
                 channel_fac=False, factored_damping=False, approx_t2=False,
                 use_float64=False, weight_decay_dict=None, max_grad_norm=0.5, verbose=1):
        """
        Kfac Optimizer for ACKTR models
        link: https://arxiv.org/pdf/1708.05144.pdf

        :param learning_rate: (float) The learning rate
        :param momentum: (float) The momentum value for the TensorFlow momentum optimizer
        :param clip_kl: (float) gradient clipping for Kullback-Leibler
        :param kfac_update: (int) update kfac after kfac_update steps
        :param stats_accum_iter: (int) how may steps to accumulate stats
        :param full_stats_init: (bool) whether or not to fully initialize stats
        :param cold_iter: (int) Cold start learning rate for how many steps
        :param cold_lr: (float) Cold start learning rate
        :param async_eigen_decomp: (bool) Use async eigen decomposition
        :param async_stats: (bool) Asynchronous stats update
        :param epsilon: (float) epsilon value for small numbers
        :param stats_decay: (float) the stats decay rate
        :param blockdiag_bias: (bool)
        :param channel_fac: (bool) factorization along the channels
        :param factored_damping: (bool) use factored damping
        :param approx_t2: (bool) approximate T2 act and grad fisher
        :param use_float64: (bool) use 64-bit float
        :param weight_decay_dict: (dict) custom weight decay coeff for a given gradient
        :param max_grad_norm: (float) The maximum value for the gradient clipping
        :param verbose: (int) verbosity level
        """
        self.max_grad_norm = max_grad_norm
        self._lr = learning_rate
        self._momentum = momentum
        self._clip_kl = clip_kl
        self._channel_fac = channel_fac
        self._kfac_update = kfac_update
        self._async_eigen_decomp = async_eigen_decomp
        self._async_stats = async_stats
        self._epsilon = epsilon
        self._stats_decay = stats_decay
        self._blockdiag_bias = blockdiag_bias
        self._approx_t2 = approx_t2
        self._use_float64 = use_float64
        self._factored_damping = factored_damping
        self._cold_iter = cold_iter
        self.verbose = verbose
        if cold_lr is None:
            # good heuristics
            self._cold_lr = self._lr  # * 3.
        else:
            self._cold_lr = cold_lr
        self._stats_accum_iter = stats_accum_iter
        if weight_decay_dict is None:
            weight_decay_dict = {}
        self._weight_decay_dict = weight_decay_dict
        self._diag_init_coeff = 0.
        self._full_stats_init = full_stats_init
        if not self._full_stats_init:
            self._stats_accum_iter = self._cold_iter

        self.sgd_step = tf.Variable(0, name='KFAC/sgd_step', trainable=False)
        self.global_step = tf.Variable(
            0, name='KFAC/global_step', trainable=False)
        self.cold_step = tf.Variable(0, name='KFAC/cold_step', trainable=False)
        self.factor_step = tf.Variable(
            0, name='KFAC/factor_step', trainable=False)
        self.stats_step = tf.Variable(
            0, name='KFAC/stats_step', trainable=False)
        self.v_f_v = tf.Variable(0., name='KFAC/vFv', trainable=False)

        self.factors = {}
        self.param_vars = []
        self.stats = {}
        self.stats_eigen = {}

        self._update_stats_op = None

    def get_factors(self, gradients, varlist):
        """
        get factors to update

        :param gradients: ([TensorFlow Tensor]) The gradients
        :param varlist: ([TensorFlow Tensor]) The parameters
        :return: ([TensorFlow Tensor]) The factors to update
        """
        default_graph = tf.get_default_graph()
        factor_tensors = {}
        fprop_tensors = []
        bprop_tensors = []
        op_types = []

        def _search_factors(gradient, graph):
            # hard coded search stratergy
            bprop_op = gradient.op
            bprop_op_name = bprop_op.name

            b_tensors = []
            f_tensors = []

            # combining additive gradient, assume they are the same op type and
            # indepedent
            if 'AddN' in bprop_op_name:
                factors = []
                for grad in gradient.op.inputs:
                    factors.append(_search_factors(grad, graph))
                op_names = [_item['opName'] for _item in factors]
                if self.verbose > 1:
                    # TODO: need to check all the attribute of the ops as well
                    print(gradient.name)
                    print(op_names)
                    print(len(np.unique(op_names)))
                assert len(np.unique(op_names)) == 1, \
                    'Error: {} is shared among different computation OPs'.format(gradient.name)

                b_tensors = reduce(lambda x, y: x + y,
                                   [_item['bpropFactors'] for _item in factors])
                if len(factors[0]['fpropFactors']) > 0:
                    f_tensors = reduce(
                        lambda x, y: x + y, [_item['fpropFactors'] for _item in factors])
                fprop_op_name = op_names[0]
                fprop_op = factors[0]['op']
            else:
                fprop_op_match = re.search('gradientsSampled(_[0-9]+|)/(.+?)_grad', bprop_op_name)
                assert fprop_op_match is not None
                fprop_op_name = fprop_op_match.group(2)
                fprop_op = graph.get_operation_by_name(fprop_op_name)
                if fprop_op.op_def.name in KFAC_OPS:
                    # Known OPs
                    b_tensor = [_i for _i in bprop_op.inputs if 'gradientsSampled' in _i.name][-1]
                    b_tensor_shape = fprop_op.outputs[0].get_shape()
                    if b_tensor.get_shape()[0].value is None:
                        b_tensor.set_shape(b_tensor_shape)
                    b_tensors.append(b_tensor)

                    if fprop_op.op_def.name == 'BiasAdd':
                        f_tensors = []
                    else:
                        f_tensors.append([_i for _i in fprop_op.inputs if param.op.name not in _i.name][0])
                    fprop_op_name = fprop_op.op_def.name
                else:
                    # unknown OPs, block approximation used
                    b_inputs_list = [_i for _i in bprop_op.inputs[0].op.inputs
                                     if 'gradientsSampled' in _i.name if 'Shape' not in _i.name]
                    if len(b_inputs_list) > 0:
                        b_tensor = b_inputs_list[0]
                        # only if tensor shape is defined, usually this will prevent tensor like Sum:0 to be used.
                        if b_tensor.get_shape():
                            b_tensor_shape = fprop_op.outputs[0].get_shape()
                            if len(b_tensor.get_shape()) > 0 and b_tensor.get_shape()[0].value is None:
                                b_tensor.set_shape(b_tensor_shape)
                            b_tensors.append(b_tensor)
                    fprop_op_name = 'UNK-' + fprop_op.op_def.name
                    op_types.append(fprop_op_name)

            return {'opName': fprop_op_name, 'op': fprop_op, 'fpropFactors': f_tensors, 'bpropFactors': b_tensors}

        for _grad, param in zip(gradients, varlist):
            if KFAC_DEBUG:
                print(('get factor for ' + param.name))
            found_factors = _search_factors(_grad, default_graph)
            factor_tensors[param] = found_factors

        # check associated weights and bias for homogeneous coordinate representation
        # and check redundent factors
        # TODO: there may be a bug to detect associate bias and weights for forking layer, e.g. in inception models.
        for param in varlist:
            factor_tensors[param]['assnWeights'] = None
            factor_tensors[param]['assnBias'] = None
        for param in varlist:
            if factor_tensors[param]['opName'] == 'BiasAdd':
                factor_tensors[param]['assnWeights'] = None
                for item in varlist:
                    if len(factor_tensors[item]['bpropFactors']) > 0:
                        if (set(factor_tensors[item]['bpropFactors']) == set(factor_tensors[param]['bpropFactors'])) \
                                and (len(factor_tensors[item]['fpropFactors']) > 0):
                            factor_tensors[param]['assnWeights'] = item
                            factor_tensors[item]['assnBias'] = param
                            factor_tensors[param]['bpropFactors'] = factor_tensors[
                                item]['bpropFactors']

        # concatenate the additive gradients along the batch dimension, i.e. assuming independence structure
        for key in ['fpropFactors', 'bpropFactors']:
            for i, param in enumerate(varlist):
                if len(factor_tensors[param][key]) > 0:
                    if (key + '_concat') not in factor_tensors[param]:
                        tensor = factor_tensors[param][key][0]  # type: tf.Tensor
                        name_scope = tensor.name.split(':')[
                            0]
                        with tf.name_scope(name_scope):
                            factor_tensors[param][
                                key + '_concat'] = tf.concat(factor_tensors[param][key], 0)
                else:
                    factor_tensors[param][key + '_concat'] = None
                for _, param2 in enumerate(varlist[(i + 1):]):
                    if (len(factor_tensors[param][key]) > 0) and (
                            set(factor_tensors[param2][key]) == set(factor_tensors[param][key])):
                        factor_tensors[param2][key] = factor_tensors[param][key]
                        factor_tensors[param2][
                            key + '_concat'] = factor_tensors[param][key + '_concat']

        if KFAC_DEBUG:
            for items in zip(varlist, fprop_tensors, bprop_tensors, op_types):
                print((items[0].name, factor_tensors[item]))
        self.factors = factor_tensors
        return factor_tensors

    def get_stats(self, factors, varlist):
        """
        return the stats values from the factors to update and the parameters

        :param factors: ([TensorFlow Tensor]) The factors to update
        :param varlist: ([TensorFlow Tensor]) The parameters
        :return: ([TensorFlow Tensor]) The stats values
        """
        if len(self.stats) == 0:
            # initialize stats variables on CPU because eigen decomp is
            # computed on CPU
            with tf.device('/cpu'):
                tmp_stats_cache = {}

                # search for tensor factors and
                # use block diag approx for the bias units
                for var in varlist:
                    bprop_factor = factors[var]['bpropFactors_concat']
                    op_type = factors[var]['opName']
                    if op_type == 'Conv2D':
                        operator_height = bprop_factor.get_shape()[1]
                        operator_width = bprop_factor.get_shape()[2]
                        if operator_height == 1 and operator_width == 1 and self._channel_fac:
                            # factorization along the channels do not support
                            # homogeneous coordinate
                            var_assn_bias = factors[var]['assnBias']
                            if var_assn_bias:
                                factors[var]['assnBias'] = None
                                factors[var_assn_bias]['assnWeights'] = None

                for var in varlist:
                    fprop_factor = factors[var]['fpropFactors_concat']
                    bprop_factor = factors[var]['bpropFactors_concat']
                    op_type = factors[var]['opName']
                    self.stats[var] = {'opName': op_type,
                                       'fprop_concat_stats': [],
                                       'bprop_concat_stats': [],
                                       'assnWeights': factors[var]['assnWeights'],
                                       'assnBias': factors[var]['assnBias'],
                                       }
                    if fprop_factor is not None:
                        if fprop_factor not in tmp_stats_cache:
                            if op_type == 'Conv2D':
                                kernel_height = var.get_shape()[0]
                                kernel_width = var.get_shape()[1]
                                n_channels = fprop_factor.get_shape()[-1]

                                operator_height = bprop_factor.get_shape()[1]
                                operator_width = bprop_factor.get_shape()[2]
                                if operator_height == 1 and operator_width == 1 and self._channel_fac:
                                    # factorization along the channels
                                    # assume independence between input channels and spatial
                                    # 2K-1 x 2K-1 covariance matrix and C x C covariance matrix
                                    # factorization along the channels do not
                                    # support homogeneous coordinate, assnBias
                                    # is always None
                                    fprop_factor2_size = kernel_height * kernel_width
                                    slot_fprop_factor_stats2 = tf.Variable(tf.diag(tf.ones(
                                        [fprop_factor2_size])) * self._diag_init_coeff,
                                                                           name='KFAC_STATS/' + fprop_factor.op.name,
                                                                           trainable=False)
                                    self.stats[var]['fprop_concat_stats'].append(
                                        slot_fprop_factor_stats2)

                                    fprop_factor_size = n_channels
                                else:
                                    # 2K-1 x 2K-1 x C x C covariance matrix
                                    # assume BHWC
                                    fprop_factor_size = kernel_height * kernel_width * n_channels
                            else:
                                # D x D covariance matrix
                                fprop_factor_size = fprop_factor.get_shape()[-1]

                            # use homogeneous coordinate
                            if not self._blockdiag_bias and self.stats[var]['assnBias']:
                                fprop_factor_size += 1

                            slot_fprop_factor_stats = tf.Variable(
                                tf.diag(tf.ones([fprop_factor_size])) * self._diag_init_coeff,
                                name='KFAC_STATS/' + fprop_factor.op.name, trainable=False)
                            self.stats[var]['fprop_concat_stats'].append(
                                slot_fprop_factor_stats)
                            if op_type != 'Conv2D':
                                tmp_stats_cache[fprop_factor] = self.stats[
                                    var]['fprop_concat_stats']
                        else:
                            self.stats[var][
                                'fprop_concat_stats'] = tmp_stats_cache[fprop_factor]

                    if bprop_factor is not None:
                        # no need to collect backward stats for bias vectors if
                        # using homogeneous coordinates
                        if not ((not self._blockdiag_bias) and self.stats[var]['assnWeights']):
                            if bprop_factor not in tmp_stats_cache:
                                slot_bprop_factor_stats = tf.Variable(tf.diag(tf.ones([bprop_factor.get_shape(
                                )[-1]])) * self._diag_init_coeff, name='KFAC_STATS/' + bprop_factor.op.name,
                                                                      trainable=False)
                                self.stats[var]['bprop_concat_stats'].append(
                                    slot_bprop_factor_stats)
                                tmp_stats_cache[bprop_factor] = self.stats[
                                    var]['bprop_concat_stats']
                            else:
                                self.stats[var][
                                    'bprop_concat_stats'] = tmp_stats_cache[bprop_factor]

        return self.stats

    def compute_and_apply_stats(self, loss_sampled, var_list=None):
        """
        compute and apply stats

        :param loss_sampled: ([TensorFlow Tensor]) the loss function output
        :param var_list: ([TensorFlow Tensor]) The parameters
        :return: (function) apply stats
        """
        varlist = var_list
        if varlist is None:
            varlist = tf.trainable_variables()

        stats = self.compute_stats(loss_sampled, var_list=varlist)
        return self.apply_stats(stats)

    def compute_stats(self, loss_sampled, var_list=None):
        """
        compute the stats values

        :param loss_sampled: ([TensorFlow Tensor]) the loss function output
        :param var_list: ([TensorFlow Tensor]) The parameters
        :return: ([TensorFlow Tensor]) stats updates
        """
        varlist = var_list
        if varlist is None:
            varlist = tf.trainable_variables()

        gradient_sampled = tf.gradients(loss_sampled, varlist, name='gradientsSampled')
        self.gradient_sampled = gradient_sampled

        # remove unused variables
        gradient_sampled, varlist = zip(*[(grad, var) for (grad, var) in zip(gradient_sampled, varlist)
                                          if grad is not None])

        factors = self.get_factors(gradient_sampled, varlist)
        stats = self.get_stats(factors, varlist)

        update_ops = []
        stats_updates = {}
        stats_updates_cache = {}
        for var in varlist:
            op_type = factors[var]['opName']
            fops = factors[var]['op']
            fprop_factor = factors[var]['fpropFactors_concat']
            fprop_stats_vars = stats[var]['fprop_concat_stats']
            bprop_factor = factors[var]['bpropFactors_concat']
            bprop_stats_vars = stats[var]['bprop_concat_stats']
            svd_factors = {}
            for stats_var in fprop_stats_vars:
                stats_var_dim = int(stats_var.get_shape()[0])
                if stats_var not in stats_updates_cache:
                    batch_size = (tf.shape(fprop_factor)[0])  # batch size
                    if op_type == 'Conv2D':
                        strides = fops.get_attr("strides")
                        padding = fops.get_attr("padding")
                        convkernel_size = var.get_shape()[0:3]

                        kernel_height = int(convkernel_size[0])
                        kernel_width = int(convkernel_size[1])
                        chan = int(convkernel_size[2])
                        flatten_size = int(kernel_height * kernel_width * chan)

                        operator_height = int(bprop_factor.get_shape()[1])
                        operator_width = int(bprop_factor.get_shape()[2])

                        if operator_height == 1 and operator_width == 1 and self._channel_fac:
                            # factorization along the channels
                            # assume independence among input channels
                            # factor = B x 1 x 1 x (KH xKW x C)
                            # patches = B x Oh x Ow x (KH xKW x C)
                            if len(svd_factors) == 0:
                                if KFAC_DEBUG:
                                    print(('approx %s act factor with rank-1 SVD factors' % var.name))
                                # find closest rank-1 approx to the feature map
                                S, U, V = tf.batch_svd(tf.reshape(
                                    fprop_factor, [-1, kernel_height * kernel_width, chan]))
                                # get rank-1 approx slides
                                sqrt_s1 = tf.expand_dims(tf.sqrt(S[:, 0, 0]), 1)
                                patches_k = U[:, :, 0] * sqrt_s1  # B x KH*KW
                                full_factor_shape = fprop_factor.get_shape()
                                patches_k.set_shape(
                                    [full_factor_shape[0], kernel_height * kernel_width])
                                patches_c = V[:, :, 0] * sqrt_s1  # B x C
                                patches_c.set_shape([full_factor_shape[0], chan])
                                svd_factors[chan] = patches_c
                                svd_factors[kernel_height * kernel_width] = patches_k
                            fprop_factor = svd_factors[stats_var_dim]

                        else:
                            # poor mem usage implementation
                            patches = tf.extract_image_patches(fprop_factor, ksizes=[1, convkernel_size[
                                0], convkernel_size[1], 1], strides=strides, rates=[1, 1, 1, 1], padding=padding)

                            if self._approx_t2:
                                if KFAC_DEBUG:
                                    print(('approxT2 act fisher for %s' % var.name))
                                # T^2 terms * 1/T^2, size: B x C
                                fprop_factor = tf.reduce_mean(patches, [1, 2])
                            else:
                                # size: (B x Oh x Ow) x C
                                fprop_factor = tf.reshape(
                                    patches, [-1, flatten_size]) / operator_height / operator_width
                    fprop_factor_size = int(fprop_factor.get_shape()[-1])
                    if stats_var_dim == (fprop_factor_size + 1) and not self._blockdiag_bias:
                        if op_type == 'Conv2D' and not self._approx_t2:
                            # correct padding for numerical stability (we
                            # divided out OhxOw from activations for T1 approx)
                            fprop_factor = tf.concat([fprop_factor, tf.ones(
                                [tf.shape(fprop_factor)[0], 1]) / operator_height / operator_width], 1)
                        else:
                            # use homogeneous coordinates
                            fprop_factor = tf.concat(
                                [fprop_factor, tf.ones([tf.shape(fprop_factor)[0], 1])], 1)

                    # average over the number of data points in a batch
                    # divided by B
                    cov = tf.matmul(fprop_factor, fprop_factor,
                                    transpose_a=True) / tf.cast(batch_size, tf.float32)
                    update_ops.append(cov)
                    stats_updates[stats_var] = cov
                    if op_type != 'Conv2D':
                        # HACK: for convolution we recompute fprop stats for
                        # every layer including forking layers
                        stats_updates_cache[stats_var] = cov

            for stats_var in bprop_stats_vars:
                if stats_var not in stats_updates_cache:
                    bprop_factor_shape = bprop_factor.get_shape()
                    batch_size = tf.shape(bprop_factor)[0]  # batch size
                    chan = int(bprop_factor_shape[-1])  # num channels
                    if op_type == 'Conv2D' or len(bprop_factor_shape) == 4:
                        if fprop_factor is not None:
                            if self._approx_t2:
                                if KFAC_DEBUG:
                                    print(('approxT2 grad fisher for %s' % var.name))
                                bprop_factor = tf.reduce_sum(
                                    bprop_factor, [1, 2])  # T^2 terms * 1/T^2
                            else:
                                bprop_factor = tf.reshape(
                                    bprop_factor, [-1, chan]) * operator_height * operator_width  # T * 1/T terms
                        else:
                            # just doing block diag approx. spatial independent
                            # structure does not apply here. summing over
                            # spatial locations
                            if KFAC_DEBUG:
                                print(('block diag approx fisher for %s' % var.name))
                            bprop_factor = tf.reduce_sum(bprop_factor, [1, 2])

                    # assume sampled loss is averaged. TODO:figure out better
                    # way to handle this
                    bprop_factor *= tf.cast(batch_size, tf.float32)
                    ##

                    cov_b = tf.matmul(bprop_factor, bprop_factor,
                                      transpose_a=True) / tf.cast(tf.shape(bprop_factor)[0], tf.float32)

                    update_ops.append(cov_b)
                    stats_updates[stats_var] = cov_b
                    stats_updates_cache[stats_var] = cov_b

        if KFAC_DEBUG:
            a_key = list(stats_updates.keys())[0]
            stats_updates[a_key] = tf.Print(stats_updates[a_key], [tf.convert_to_tensor('step:'), self.global_step,
                                                                   tf.convert_to_tensor('computing stats')])
        self.stats_updates = stats_updates
        return stats_updates

    def apply_stats(self, stats_updates):
        """
        compute stats and update/apply the new stats to the running average

        :param stats_updates: ([TensorFlow Tensor]) The stats updates
        :return: (function) update stats operation
        """

        def _update_accum_stats():
            if self._full_stats_init:
                return tf.cond(tf.greater(self.sgd_step, self._cold_iter), lambda: tf.group(
                    *self._apply_stats(stats_updates, accumulate=True, accumulate_coeff=1. / self._stats_accum_iter)),
                               tf.no_op)
            else:
                return tf.group(
                    *self._apply_stats(stats_updates, accumulate=True, accumulate_coeff=1. / self._stats_accum_iter))

        def _update_running_avg_stats(stats_updates):
            return tf.group(*self._apply_stats(stats_updates))

        if self._async_stats:
            # asynchronous stats update
            update_stats = self._apply_stats(stats_updates)

            queue = tf.FIFOQueue(1, [item.dtype for item in update_stats], shapes=[
                item.get_shape() for item in update_stats])
            enqueue_op = queue.enqueue(update_stats)

            def dequeue_stats_op():
                return queue.dequeue()

            self.qr_stats = tf.train.QueueRunner(queue, [enqueue_op])
            update_stats_op = tf.cond(tf.equal(queue.size(), tf.convert_to_tensor(
                0)), tf.no_op, lambda: tf.group(*[dequeue_stats_op(), ]))
        else:
            # synchronous stats update
            update_stats_op = tf.cond(tf.greater_equal(self.stats_step, self._stats_accum_iter),
                                      lambda: _update_running_avg_stats(stats_updates), _update_accum_stats)
        self._update_stats_op = update_stats_op
        return update_stats_op

    def _apply_stats(self, stats_updates, accumulate=False, accumulate_coeff=0.):
        update_ops = []
        # obtain the stats var list
        for stats_var in stats_updates:
            stats_new = stats_updates[stats_var]
            if accumulate:
                # simple superbatch averaging
                update_op = tf.assign_add(
                    stats_var, accumulate_coeff * stats_new, use_locking=True)
            else:
                # exponential running averaging
                update_op = tf.assign(
                    stats_var, stats_var * self._stats_decay, use_locking=True)
                update_op = tf.assign_add(
                    update_op, (1. - self._stats_decay) * stats_new, use_locking=True)
            update_ops.append(update_op)

        with tf.control_dependencies(update_ops):
            stats_step_op = tf.assign_add(self.stats_step, 1)

        if KFAC_DEBUG:
            stats_step_op = (tf.Print(stats_step_op,
                                      [tf.convert_to_tensor('step:'),
                                       self.global_step,
                                       tf.convert_to_tensor('fac step:'),
                                       self.factor_step,
                                       tf.convert_to_tensor('sgd step:'),
                                       self.sgd_step,
                                       tf.convert_to_tensor('Accum:'),
                                       tf.convert_to_tensor(accumulate),
                                       tf.convert_to_tensor('Accum coeff:'),
                                       tf.convert_to_tensor(accumulate_coeff),
                                       tf.convert_to_tensor('stat step:'),
                                       self.stats_step, update_ops[0], update_ops[1]]))
        return [stats_step_op, ]

    def get_stats_eigen(self, stats=None):
        """
        Return the eigen values from the stats

        :param stats: ([TensorFlow Tensor]) The stats
        :return: ([TensorFlow Tensor]) The stats eigen values
        """
        if len(self.stats_eigen) == 0:
            stats_eigen = {}
            if stats is None:
                stats = self.stats

            tmp_eigen_cache = {}
            with tf.device('/cpu:0'):
                for var in stats:
                    for key in ['fprop_concat_stats', 'bprop_concat_stats']:
                        for stats_var in stats[var][key]:
                            if stats_var not in tmp_eigen_cache:
                                stats_dim = stats_var.get_shape()[1].value
                                eigen_values = tf.Variable(tf.ones(
                                    [stats_dim]), name='KFAC_FAC/' + stats_var.name.split(':')[0] + '/e',
                                    trainable=False)
                                eigen_vectors = tf.Variable(tf.diag(tf.ones(
                                    [stats_dim])), name='KFAC_FAC/' + stats_var.name.split(':')[0] + '/Q',
                                    trainable=False)
                                stats_eigen[stats_var] = {'e': eigen_values, 'Q': eigen_vectors}
                                tmp_eigen_cache[
                                    stats_var] = stats_eigen[stats_var]
                            else:
                                stats_eigen[stats_var] = tmp_eigen_cache[
                                    stats_var]
            self.stats_eigen = stats_eigen
        return self.stats_eigen

    def compute_stats_eigen(self):
        """
        compute the eigen decomp using copied var stats to avoid concurrent read/write from other queue

        :return: ([TensorFlow Tensor]) update operations
        """
        # TODO: figure out why this op has delays (possibly moving eigenvectors around?)
        with tf.device('/cpu:0'):
            stats_eigen = self.stats_eigen
            computed_eigen = {}
            eigen_reverse_lookup = {}
            update_ops = []
            # sync copied stats
            with tf.control_dependencies([]):
                for stats_var in stats_eigen:
                    if stats_var not in computed_eigen:
                        eigen_decomposition = tf.self_adjoint_eig(stats_var)
                        eigen_values = eigen_decomposition[0]
                        eigen_vectors = eigen_decomposition[1]
                        if self._use_float64:
                            eigen_values = tf.cast(eigen_values, tf.float64)
                            eigen_vectors = tf.cast(eigen_vectors, tf.float64)
                        update_ops.append(eigen_values)
                        update_ops.append(eigen_vectors)
                        computed_eigen[stats_var] = {'e': eigen_values, 'Q': eigen_vectors}
                        eigen_reverse_lookup[eigen_values] = stats_eigen[stats_var]['e']
                        eigen_reverse_lookup[eigen_vectors] = stats_eigen[stats_var]['Q']

            self.eigen_reverse_lookup = eigen_reverse_lookup
            self.eigen_update_list = update_ops

            if KFAC_DEBUG:
                self.eigen_update_list = [item for item in update_ops]
                with tf.control_dependencies(update_ops):
                    update_ops.append(tf.Print(tf.constant(
                        0.), [tf.convert_to_tensor('computed factor eigen')]))

        return update_ops

    def apply_stats_eigen(self, eigen_list):
        """
        apply the update using the eigen values of the stats

        :param eigen_list: ([TensorFlow Tensor]) The list of eigen values of the stats
        :return: ([TensorFlow Tensor]) update operations
        """
        update_ops = []
        if self.verbose > 1:
            print(('updating %d eigenvalue/vectors' % len(eigen_list)))
        for _, (tensor, mark) in enumerate(zip(eigen_list, self.eigen_update_list)):
            stats_eigen_var = self.eigen_reverse_lookup[mark]
            update_ops.append(
                tf.assign(stats_eigen_var, tensor, use_locking=True))

        with tf.control_dependencies(update_ops):
            factor_step_op = tf.assign_add(self.factor_step, 1)
            update_ops.append(factor_step_op)
            if KFAC_DEBUG:
                update_ops.append(tf.Print(tf.constant(
                    0.), [tf.convert_to_tensor('updated kfac factors')]))
        return update_ops

    def get_kfac_precond_updates(self, gradlist, varlist):
        """
        return the KFAC updates

        :param gradlist: ([TensorFlow Tensor]) The gradients
        :param varlist: ([TensorFlow Tensor]) The parameters
        :return: ([TensorFlow Tensor]) the update list
        """
        v_g = 0.

        assert len(self.stats) > 0
        assert len(self.stats_eigen) > 0
        assert len(self.factors) > 0
        counter = 0

        grad_dict = {var: grad for grad, var in zip(gradlist, varlist)}

        for grad, var in zip(gradlist, varlist):
            grad_reshape = False

            fprop_factored_fishers = self.stats[var]['fprop_concat_stats']
            bprop_factored_fishers = self.stats[var]['bprop_concat_stats']

            if (len(fprop_factored_fishers) + len(bprop_factored_fishers)) > 0:
                counter += 1
                grad_shape = grad.get_shape()
                if len(grad.get_shape()) > 2:
                    # reshape conv kernel parameters
                    kernel_width = int(grad.get_shape()[0])
                    kernel_height = int(grad.get_shape()[1])
                    n_channels = int(grad.get_shape()[2])
                    depth = int(grad.get_shape()[3])

                    if len(fprop_factored_fishers) > 1 and self._channel_fac:
                        # reshape conv kernel parameters into tensor
                        grad = tf.reshape(grad, [kernel_width * kernel_height, n_channels, depth])
                    else:
                        # reshape conv kernel parameters into 2D grad
                        grad = tf.reshape(grad, [-1, depth])
                    grad_reshape = True
                elif len(grad.get_shape()) == 1:
                    # reshape bias or 1D parameters

                    grad = tf.expand_dims(grad, 0)
                    grad_reshape = True

                if (self.stats[var]['assnBias'] is not None) and not self._blockdiag_bias:
                    # use homogeneous coordinates only works for 2D grad.
                    # TODO: figure out how to factorize bias grad
                    # stack bias grad
                    var_assn_bias = self.stats[var]['assnBias']
                    grad = tf.concat(
                        [grad, tf.expand_dims(grad_dict[var_assn_bias], 0)], 0)

                # project gradient to eigen space and reshape the eigenvalues
                # for broadcasting
                eig_vals = []

                for idx, stats in enumerate(self.stats[var]['fprop_concat_stats']):
                    eigen_vectors = self.stats_eigen[stats]['Q']
                    eigen_values = detect_min_val(self.stats_eigen[stats][
                                                      'e'], var, name='act', debug=KFAC_DEBUG)

                    eigen_vectors, eigen_values = factor_reshape(eigen_vectors, eigen_values,
                                                                 grad, fac_idx=idx, f_type='act')
                    eig_vals.append(eigen_values)
                    grad = gmatmul(eigen_vectors, grad, transpose_a=True, reduce_dim=idx)

                for idx, stats in enumerate(self.stats[var]['bprop_concat_stats']):
                    eigen_vectors = self.stats_eigen[stats]['Q']
                    eigen_values = detect_min_val(self.stats_eigen[stats][
                                                      'e'], var, name='grad', debug=KFAC_DEBUG)

                    eigen_vectors, eigen_values = factor_reshape(eigen_vectors, eigen_values,
                                                                 grad, fac_idx=idx, f_type='grad')
                    eig_vals.append(eigen_values)
                    grad = gmatmul(grad, eigen_vectors, transpose_b=False, reduce_dim=idx)

                # whiten using eigenvalues
                weight_decay_coeff = 0.
                if var in self._weight_decay_dict:
                    weight_decay_coeff = self._weight_decay_dict[var]
                    if KFAC_DEBUG:
                        print(('weight decay coeff for %s is %f' % (var.name, weight_decay_coeff)))

                if self._factored_damping:
                    if KFAC_DEBUG:
                        print(('use factored damping for %s' % var.name))
                    coeffs = 1.
                    num_factors = len(eig_vals)
                    # compute the ratio of two trace norm of the left and right
                    # KFac matrices, and their generalization
                    if len(eig_vals) == 1:
                        damping = self._epsilon + weight_decay_coeff
                    else:
                        damping = tf.pow(
                            self._epsilon + weight_decay_coeff, 1. / num_factors)
                    eig_vals_tnorm_avg = [tf.reduce_mean(
                        tf.abs(e)) for e in eig_vals]
                    for eigen_val, e_tnorm in zip(eig_vals, eig_vals_tnorm_avg):
                        eig_tnorm_neg_list = [
                            item for item in eig_vals_tnorm_avg if item != e_tnorm]
                        if len(eig_vals) == 1:
                            adjustment = 1.
                        elif len(eig_vals) == 2:
                            adjustment = tf.sqrt(
                                e_tnorm / eig_tnorm_neg_list[0])
                        else:
                            eig_tnorm_neg_list_prod = reduce(
                                lambda x, y: x * y, eig_tnorm_neg_list)
                            adjustment = tf.pow(
                                tf.pow(e_tnorm, num_factors - 1.) / eig_tnorm_neg_list_prod, 1. / num_factors)
                        coeffs *= (eigen_val + adjustment * damping)
                else:
                    coeffs = 1.
                    damping = (self._epsilon + weight_decay_coeff)
                    for eigen_val in eig_vals:
                        coeffs *= eigen_val
                    coeffs += damping

                grad /= coeffs

                # project gradient back to euclidean space
                for idx, stats in enumerate(self.stats[var]['fprop_concat_stats']):
                    eigen_vectors = self.stats_eigen[stats]['Q']
                    grad = gmatmul(eigen_vectors, grad, transpose_a=False, reduce_dim=idx)

                for idx, stats in enumerate(self.stats[var]['bprop_concat_stats']):
                    eigen_vectors = self.stats_eigen[stats]['Q']
                    grad = gmatmul(grad, eigen_vectors, transpose_b=True, reduce_dim=idx)

                if (self.stats[var]['assnBias'] is not None) and not self._blockdiag_bias:
                    # use homogeneous coordinates only works for 2D grad.
                    # TODO: figure out how to factorize bias grad
                    # un-stack bias grad
                    var_assn_bias = self.stats[var]['assnBias']
                    c_plus_one = int(grad.get_shape()[0])
                    grad_assn_bias = tf.reshape(tf.slice(grad,
                                                         begin=[
                                                             c_plus_one - 1, 0],
                                                         size=[1, -1]), var_assn_bias.get_shape())
                    grad_assn_weights = tf.slice(grad,
                                                 begin=[0, 0],
                                                 size=[c_plus_one - 1, -1])
                    grad_dict[var_assn_bias] = grad_assn_bias
                    grad = grad_assn_weights

                if grad_reshape:
                    grad = tf.reshape(grad, grad_shape)

                grad_dict[var] = grad

        if self.verbose > 1:
            print(('projecting %d gradient matrices' % counter))

        for grad_1, var in zip(gradlist, varlist):
            grad = grad_dict[var]
            # clipping
            if KFAC_DEBUG:
                print(('apply clipping to %s' % var.name))
                tf.Print(grad, [tf.sqrt(tf.reduce_sum(tf.pow(grad, 2)))], "Euclidean norm of new grad")
            local_vg = tf.reduce_sum(grad * grad_1 * (self._lr * self._lr))
            v_g += local_vg

        # rescale everything
        if KFAC_DEBUG:
            print('apply vFv clipping')

        scaling = tf.minimum(1., tf.sqrt(self._clip_kl / v_g))
        if KFAC_DEBUG:
            scaling = tf.Print(scaling, [tf.convert_to_tensor(
                'clip: '), scaling, tf.convert_to_tensor(' vFv: '), v_g])
        with tf.control_dependencies([tf.assign(self.v_f_v, v_g)]):
            updatelist = [grad_dict[var] for var in varlist]
            for i, item in enumerate(updatelist):
                updatelist[i] = scaling * item

        return updatelist

    @classmethod
    def compute_gradients(cls, loss, var_list=None):
        """
        compute the gradients from the loss and the parameters

        :param loss: ([TensorFlow Tensor]) The loss
        :param var_list: ([TensorFlow Tensor]) The parameters
        :return: ([TensorFlow Tensor]) the gradient
        """
        varlist = var_list
        if varlist is None:
            varlist = tf.trainable_variables()
        gradients = tf.gradients(loss, varlist)

        return [(a, b) for a, b in zip(gradients, varlist)]

    def apply_gradients_kfac(self, grads):
        """
        apply the kfac gradient

        :param grads: ([TensorFlow Tensor]) the gradient
        :return: ([function], QueueRunner) Update functions, queue operation runner
        """
        grad, varlist = list(zip(*grads))

        if len(self.stats_eigen) == 0:
            self.get_stats_eigen()

        queue_runner = None
        # launch eigen-decomp on a queue thread
        if self._async_eigen_decomp:
            if self.verbose > 1:
                print('Using async eigen decomposition')
            # get a list of factor loading tensors
            factor_ops_dummy = self.compute_stats_eigen()

            # define a queue for the list of factor loading tensors
            queue = tf.FIFOQueue(1, [item.dtype for item in factor_ops_dummy],
                                 shapes=[item.get_shape() for item in factor_ops_dummy])
            enqueue_op = tf.cond(
                tf.logical_and(tf.equal(tf.mod(self.stats_step, self._kfac_update), tf.convert_to_tensor(
                    0)), tf.greater_equal(self.stats_step, self._stats_accum_iter)),
                lambda: queue.enqueue(self.compute_stats_eigen()), tf.no_op)

            def dequeue_op():
                return queue.dequeue()

            queue_runner = tf.train.QueueRunner(queue, [enqueue_op])

        update_ops = []
        global_step_op = tf.assign_add(self.global_step, 1)
        update_ops.append(global_step_op)

        with tf.control_dependencies([global_step_op]):

            # compute updates
            assert self._update_stats_op is not None
            update_ops.append(self._update_stats_op)
            dependency_list = []
            if not self._async_eigen_decomp:
                dependency_list.append(self._update_stats_op)

            with tf.control_dependencies(dependency_list):
                def no_op_wrapper():
                    return tf.group(*[tf.assign_add(self.cold_step, 1)])

                if not self._async_eigen_decomp:
                    # synchronous eigen-decomp updates
                    update_factor_ops = tf.cond(tf.logical_and(tf.equal(tf.mod(self.stats_step, self._kfac_update),
                                                                        tf.convert_to_tensor(0)),
                                                               tf.greater_equal(self.stats_step,
                                                                                self._stats_accum_iter)),
                                                lambda: tf.group(*self.apply_stats_eigen(self.compute_stats_eigen())),
                                                no_op_wrapper)
                else:
                    # asynchronous eigen-decomp updates using queue
                    update_factor_ops = tf.cond(tf.greater_equal(self.stats_step, self._stats_accum_iter),
                                                lambda: tf.cond(tf.equal(queue.size(), tf.convert_to_tensor(0)),
                                                                tf.no_op,

                                                                lambda: tf.group(
                                                                    *self.apply_stats_eigen(dequeue_op())),
                                                                ),
                                                no_op_wrapper)

                update_ops.append(update_factor_ops)

                with tf.control_dependencies([update_factor_ops]):
                    def grad_op():
                        return list(grad)

                    def get_kfac_grad_op():
                        return self.get_kfac_precond_updates(grad, varlist)

                    u = tf.cond(tf.greater(self.factor_step,
                                           tf.convert_to_tensor(0)), get_kfac_grad_op, grad_op)

                    optim = tf.train.MomentumOptimizer(
                        self._lr * (1. - self._momentum), self._momentum)

                    # optim = tf.train.AdamOptimizer(self._lr, epsilon=0.01)

                    def optim_op():
                        def update_optim_op():
                            if self._full_stats_init:
                                return tf.cond(tf.greater(self.factor_step, tf.convert_to_tensor(0)),
                                               lambda: optim.apply_gradients(list(zip(u, varlist))), tf.no_op)
                            else:
                                return optim.apply_gradients(list(zip(u, varlist)))

                        if self._full_stats_init:
                            return tf.cond(tf.greater_equal(self.stats_step, self._stats_accum_iter), update_optim_op,
                                           tf.no_op)
                        else:
                            return tf.cond(tf.greater_equal(self.sgd_step, self._cold_iter), update_optim_op, tf.no_op)

                    update_ops.append(optim_op())

        return tf.group(*update_ops), queue_runner

    def apply_gradients(self, grads):
        """
        apply the gradient

        :param grads: ([TensorFlow Tensor]) the gradient
        :return: (function, QueueRunner) train operation, queue operation runner
        """
        cold_optim = tf.train.MomentumOptimizer(self._cold_lr, self._momentum)

        def _cold_sgd_start():
            sgd_grads, sgd_var = zip(*grads)

            if self.max_grad_norm is not None:
                sgd_grads, _ = tf.clip_by_global_norm(sgd_grads, self.max_grad_norm)

            sgd_grads = list(zip(sgd_grads, sgd_var))

            sgd_step_op = tf.assign_add(self.sgd_step, 1)
            cold_optim_op = cold_optim.apply_gradients(sgd_grads)
            if KFAC_DEBUG:
                with tf.control_dependencies([sgd_step_op, cold_optim_op]):
                    sgd_step_op = tf.Print(
                        sgd_step_op, [self.sgd_step, tf.convert_to_tensor('doing cold sgd step')])
            return tf.group(*[sgd_step_op, cold_optim_op])

        # remove unused variables
        grads = [(grad, var) for (grad, var) in grads if grad is not None]

        kfac_optim_op, queue_runner = self.apply_gradients_kfac(grads)

        def _warm_kfac_start():
            return kfac_optim_op

        return tf.cond(tf.greater(self.sgd_step, self._cold_iter), _warm_kfac_start, _cold_sgd_start), queue_runner

    def minimize(self, loss, loss_sampled, var_list=None):
        """
        minimize the gradient loss

        :param loss: ([TensorFlow Tensor]) The loss
        :param loss_sampled: ([TensorFlow Tensor]) the loss function output
        :param var_list: ([TensorFlow Tensor]) The parameters
        :return: (function, q_runner) train operation, queue operation runner
        """
        grads = self.compute_gradients(loss, var_list=var_list)
        self.compute_and_apply_stats(loss_sampled, var_list=var_list)
        return self.apply_gradients(grads)



from neorl import FNEAT
import numpy as np 

def test_fneat():
 
    def Sphere(individual):
        """Sphere test objective function.
                F(x) = sum_{i=1}^d xi^2
                d=1,2,3,...
                Range: [-100,100]
                Minima: 0
        """
        return sum(x**2 for x in individual)
    
    nx=5
    lb=-100
    ub=100
    bounds={}
    for i in range(1,nx+1):
            bounds['x'+str(i)]=['float', -100, 100]
    
    # modify your own NEAT config
    config = {
        'pop_size': 50,
        'num_hidden': 1,
        'activation_mutate_rate': 0.1,
        'survival_threshold': 0.3,
        }
    
    fneat=FNEAT(fit=Sphere, bounds=bounds, mode='min', config= config, ncores=1, seed=1)
    #some random guess (just one individual)
    x0 = np.random.uniform(lb,ub,nx)
    x_best, y_best, fneat_hist=fneat.evolute(ngen=5, x0=x0, verbose=True, 
                                              checkpoint_itv=None, startpoint=None) 
    
test_fneat()


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# ------------------------------------------------------------------------
# This code is taken from the following paper:
#
# Pengfei Huang,Handing Wang,Yaochu Jin,Offine Data-Driven Evolutionary Optimization Based on Tri-Training, Swarm and Evolutionary Computation, Accepted.
#
# You are free to use it for non-commercial purposes. However, we do not offer any forms of guanrantee or warranty associated with the code. We would appreciate your acknowledgement.
# ------------------------------------------------------------------------
import numpy as np

def latin(N, D, lower_bound, upper_bound):
    d = 1.0 / N
    result = np.empty([N, D])
    temp = np.empty([N])
    for i in range(D):
        for j in range(N):
            temp[j] = np.random.uniform(
                low=j * d, high=(j + 1) * d, size=1)[0]
        np.random.shuffle(temp)
        for j in range(N):
            result[j, i] = temp[j]
    if np.any(lower_bound > upper_bound):
        print('Range error')
        return None
    np.add(np.multiply(result, (upper_bound - lower_bound), out=result), lower_bound, out=result)
    return result


import random
from typing import Optional, List, Union

import numpy as np

from neorl.rl.baselines.shared.segment_tree import SumSegmentTree, MinSegmentTree
from neorl.rl.baselines.shared.vec_env import VecNormalize


class ReplayBuffer(object):
    def __init__(self, size: int):
        """
        Implements a ring buffer (FIFO).

        :param size: (int)  Max number of transitions to store in the buffer. When the buffer overflows the old
            memories are dropped.
        """
        self._storage = []
        self._maxsize = size
        self._next_idx = 0

    def __len__(self) -> int:
        return len(self._storage)

    @property
    def storage(self):
        """[(Union[np.ndarray, int], Union[np.ndarray, int], float, Union[np.ndarray, int], bool)]: content of the replay buffer"""
        return self._storage

    @property
    def buffer_size(self) -> int:
        """float: Max capacity of the buffer"""
        return self._maxsize

    def can_sample(self, n_samples: int) -> bool:
        """
        Check if n_samples samples can be sampled
        from the buffer.

        :param n_samples: (int)
        :return: (bool)
        """
        return len(self) >= n_samples

    def is_full(self) -> int:
        """
        Check whether the replay buffer is full or not.

        :return: (bool)
        """
        return len(self) == self.buffer_size

    def add(self, obs_t, action, reward, obs_tp1, done):
        """
        add a new transition to the buffer

        :param obs_t: (Union[np.ndarray, int]) the last observation
        :param action: (Union[np.ndarray, int]) the action
        :param reward: (float) the reward of the transition
        :param obs_tp1: (Union[np.ndarray, int]) the current observation
        :param done: (bool) is the episode done
        """
        data = (obs_t, action, reward, obs_tp1, done)

        if self._next_idx >= len(self._storage):
            self._storage.append(data)
        else:
            self._storage[self._next_idx] = data
        self._next_idx = (self._next_idx + 1) % self._maxsize

    def extend(self, obs_t, action, reward, obs_tp1, done):
        """
        add a new batch of transitions to the buffer

        :param obs_t: (Union[Tuple[Union[np.ndarray, int]], np.ndarray]) the last batch of observations
        :param action: (Union[Tuple[Union[np.ndarray, int]]], np.ndarray]) the batch of actions
        :param reward: (Union[Tuple[float], np.ndarray]) the batch of the rewards of the transition
        :param obs_tp1: (Union[Tuple[Union[np.ndarray, int]], np.ndarray]) the current batch of observations
        :param done: (Union[Tuple[bool], np.ndarray]) terminal status of the batch

        Note: uses the same names as .add to keep compatibility with named argument passing
                but expects iterables and arrays with more than 1 dimensions
        """
        for data in zip(obs_t, action, reward, obs_tp1, done):
            if self._next_idx >= len(self._storage):
                self._storage.append(data)
            else:
                self._storage[self._next_idx] = data
            self._next_idx = (self._next_idx + 1) % self._maxsize

    @staticmethod
    def _normalize_obs(obs: np.ndarray,
                       env: Optional[VecNormalize] = None) -> np.ndarray:
        """
        Helper for normalizing the observation.
        """
        if env is not None:
            return env.normalize_obs(obs)
        return obs

    @staticmethod
    def _normalize_reward(reward: np.ndarray,
                          env: Optional[VecNormalize] = None) -> np.ndarray:
        """
        Helper for normalizing the reward.
        """
        if env is not None:
            return env.normalize_reward(reward)
        return reward

    def _encode_sample(self, idxes: Union[List[int], np.ndarray], env: Optional[VecNormalize] = None):
        obses_t, actions, rewards, obses_tp1, dones = [], [], [], [], []
        for i in idxes:
            data = self._storage[i]
            obs_t, action, reward, obs_tp1, done = data
            obses_t.append(np.array(obs_t, copy=False))
            actions.append(np.array(action, copy=False))
            rewards.append(reward)
            obses_tp1.append(np.array(obs_tp1, copy=False))
            dones.append(done)
        return (self._normalize_obs(np.array(obses_t), env),
                np.array(actions),
                self._normalize_reward(np.array(rewards), env),
                self._normalize_obs(np.array(obses_tp1), env),
                np.array(dones))

    def sample(self, batch_size: int, env: Optional[VecNormalize] = None, **_kwargs):
        """
        Sample a batch of experiences.

        :param batch_size: (int) How many transitions to sample.
        :param env: (Optional[VecNormalize]) associated gym VecEnv
            to normalize the observations/rewards when sampling
        :return:
            - obs_batch: (np.ndarray) batch of observations
            - act_batch: (numpy float) batch of actions executed given obs_batch
            - rew_batch: (numpy float) rewards received as results of executing act_batch
            - next_obs_batch: (np.ndarray) next set of observations seen after executing act_batch
            - done_mask: (numpy bool) done_mask[i] = 1 if executing act_batch[i] resulted in the end of an episode
                and 0 otherwise.
        """
        idxes = [random.randint(0, len(self._storage) - 1) for _ in range(batch_size)]
        return self._encode_sample(idxes, env=env)


class PrioritizedReplayBuffer(ReplayBuffer):
    def __init__(self, size, alpha):
        """
        Create Prioritized Replay buffer.

        See Also ReplayBuffer.__init__

        :param size: (int) Max number of transitions to store in the buffer. When the buffer overflows the old memories
            are dropped.
        :param alpha: (float) how much prioritization is used (0 - no prioritization, 1 - full prioritization)
        """
        super(PrioritizedReplayBuffer, self).__init__(size)
        assert alpha >= 0
        self._alpha = alpha

        it_capacity = 1
        while it_capacity < size:
            it_capacity *= 2

        self._it_sum = SumSegmentTree(it_capacity)
        self._it_min = MinSegmentTree(it_capacity)
        self._max_priority = 1.0

    def add(self, obs_t, action, reward, obs_tp1, done):
        """
        add a new transition to the buffer

        :param obs_t: (Any) the last observation
        :param action: ([float]) the action
        :param reward: (float) the reward of the transition
        :param obs_tp1: (Any) the current observation
        :param done: (bool) is the episode done
        """
        idx = self._next_idx
        super().add(obs_t, action, reward, obs_tp1, done)
        self._it_sum[idx] = self._max_priority ** self._alpha
        self._it_min[idx] = self._max_priority ** self._alpha

    def extend(self, obs_t, action, reward, obs_tp1, done):
        """
        add a new batch of transitions to the buffer

        :param obs_t: (Union[Tuple[Union[np.ndarray, int]], np.ndarray]) the last batch of observations
        :param action: (Union[Tuple[Union[np.ndarray, int]]], np.ndarray]) the batch of actions
        :param reward: (Union[Tuple[float], np.ndarray]) the batch of the rewards of the transition
        :param obs_tp1: (Union[Tuple[Union[np.ndarray, int]], np.ndarray]) the current batch of observations
        :param done: (Union[Tuple[bool], np.ndarray]) terminal status of the batch

        Note: uses the same names as .add to keep compatibility with named argument passing
            but expects iterables and arrays with more than 1 dimensions
        """
        idx = self._next_idx
        super().extend(obs_t, action, reward, obs_tp1, done)
        while idx != self._next_idx:
            self._it_sum[idx] = self._max_priority ** self._alpha
            self._it_min[idx] = self._max_priority ** self._alpha
            idx = (idx + 1) % self._maxsize

    def _sample_proportional(self, batch_size):
        mass = []
        total = self._it_sum.sum(0, len(self._storage) - 1)
        # TODO(szymon): should we ensure no repeats?
        mass = np.random.random(size=batch_size) * total
        idx = self._it_sum.find_prefixsum_idx(mass)
        return idx

    def sample(self, batch_size: int, beta: float = 0, env: Optional[VecNormalize] = None):
        """
        Sample a batch of experiences.

        compared to ReplayBuffer.sample
        it also returns importance weights and idxes
        of sampled experiences.

        :param batch_size: (int) How many transitions to sample.
        :param beta: (float) To what degree to use importance weights (0 - no corrections, 1 - full correction)
        :param env: (Optional[VecNormalize]) associated gym VecEnv
            to normalize the observations/rewards when sampling
        :return:
            - obs_batch: (np.ndarray) batch of observations
            - act_batch: (numpy float) batch of actions executed given obs_batch
            - rew_batch: (numpy float) rewards received as results of executing act_batch
            - next_obs_batch: (np.ndarray) next set of observations seen after executing act_batch
            - done_mask: (numpy bool) done_mask[i] = 1 if executing act_batch[i] resulted in the end of an episode
                and 0 otherwise.
            - weights: (numpy float) Array of shape (batch_size,) and dtype np.float32 denoting importance weight of
                each sampled transition
            - idxes: (numpy int) Array of shape (batch_size,) and dtype np.int32 idexes in buffer of sampled experiences
        """
        assert beta > 0

        idxes = self._sample_proportional(batch_size)
        weights = []
        p_min = self._it_min.min() / self._it_sum.sum()
        max_weight = (p_min * len(self._storage)) ** (-beta)
        p_sample = self._it_sum[idxes] / self._it_sum.sum()
        weights = (p_sample * len(self._storage)) ** (-beta) / max_weight
        encoded_sample = self._encode_sample(idxes, env=env)
        return tuple(list(encoded_sample) + [weights, idxes])

    def update_priorities(self, idxes, priorities):
        """
        Update priorities of sampled transitions.

        sets priority of transition at index idxes[i] in buffer
        to priorities[i].

        :param idxes: ([int]) List of idxes of sampled transitions
        :param priorities: ([float]) List of updated priorities corresponding to transitions at the sampled idxes
            denoted by variable `idxes`.
        """
        assert len(idxes) == len(priorities)
        assert np.min(priorities) > 0
        assert np.min(idxes) >= 0
        assert np.max(idxes) < len(self.storage)
        self._it_sum[idxes] = priorities ** self._alpha
        self._it_min[idxes] = priorities ** self._alpha

        self._max_priority = max(self._max_priority, np.max(priorities))



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Thu Dec  3 14:42:29 2020
#
#@author: Majdi
#"""


import random
import numpy as np
import math
#from solution import solution
import time
from collections import defaultdict
import sys
import uuid

import multiprocessing
import multiprocessing.pool
from neorl.evolu.discrete import mutate_discrete
from neorl.utils.seeding import set_neorl_seed

class NoDaemonProcess(multiprocessing.Process):
    # make 'daemon' attribute always return False
    def _get_daemon(self):
        return False
    def _set_daemon(self, value):
        pass
    daemon = property(_get_daemon, _set_daemon)

# We sub-class multiprocessing.pool.Pool instead of multiprocessing.Pool
# because the latter is only a wrapper function, not a proper class.
class MyPool(multiprocessing.pool.Pool):
    Process = NoDaemonProcess

#multiprocessing trick to paralllelize nested functions in python (un-picklable objects!)
def globalize(func):
  def result(*args, **kwargs):
    return -func(*args, **kwargs)
  result.__name__ = result.__qualname__ = uuid.uuid4().hex
  setattr(sys.modules[result.__module__], result.__name__, result)
  return result
    
class GWOmod(object):
    """
    Grey Wolf Optimizer
    
    :param mode: (str) problem type, either "min" for minimization problem or "max" for maximization
    :param bounds: (dict) input parameter type and lower/upper bounds in dictionary form. Example: {'x1': ['int', 1, 4], 'x2': ['float', 0.1, 0.8], 'x3': ['float', 2.2, 6.2]}
    :param fit: (function) the fitness function 
    :param nwolves: (int): number of the grey wolves in the group
    :param ncores: (int) number of parallel processors (must be ``<= nwolves``)
    :param seed: (int) random seed for sampling
    """
    def __init__(self, mode, bounds, fit, nwolves=5, int_transform ='nearest_int', ncores=1, seed=None):
        
        
        set_neorl_seed(seed)
        
#        #--mir
        self.mode=mode
        if mode == 'min':
            self.fit=fit
        elif mode == 'max':
            self.fit=globalize(lambda x: fit(x))  #use the function globalize to serialize the nested fit
        else:
            raise ValueError('--error: The mode entered by user is invalid, use either `min` or `max`')
            
        self.bounds=bounds
        self.ncores = ncores
        self.nwolves=nwolves
        self.dim = len(bounds)
        self.int_transform=int_transform
        
        self.var_type = np.array([bounds[item][0] for item in bounds])
        self.lb=[self.bounds[item][1] for item in self.bounds]
        self.ub=[self.bounds[item][2] for item in self.bounds]

    def init_sample(self, bounds):
    
        indv=[]
        for key in bounds:
            if bounds[key][0] == 'int':
                indv.append(random.randint(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'float':
                indv.append(random.uniform(bounds[key][1], bounds[key][2]))
            #elif bounds[key][0] == 'grid':
            #    indv.append(random.sample(bounds[key][1],1)[0])
            else:
                raise Exception ('unknown data type is given, either int, float, or grid are allowed for parameter bounds')   
        return indv
    
    def ensure_bounds(self, vec):
    
        vec_new = []
        # cycle through each variable in vector 
        for i, (key, val) in enumerate(self.bounds.items()):
    
            # variable exceedes the minimum boundary
            if vec[i] < self.bounds[key][1]:
                vec_new.append(self.bounds[key][1])
    
            # variable exceedes the maximum boundary
            if vec[i] > self.bounds[key][2]:
                vec_new.append(self.bounds[key][2])
    
            # the variable is fine
            if self.bounds[key][1] <= vec[i] <= self.bounds[key][2]:
                vec_new.append(vec[i])
            
        return vec_new
    
    def ensure_discrete(self, vec):
        #"""
        #to mutate a vector if discrete variables exist 

        #Params:
        #vec - position in vector/list form

        #Return:
        #vec - updated position vector with discrete values
        #"""
        
        for dim in range(self.dim):
            if self.var_type[dim] == 'int':
                vec[dim] = mutate_discrete(x_ij=vec[dim], 
                                               x_min=min(vec),
                                               x_max=max(vec),
                                               lb=self.lb[dim], 
                                               ub=self.ub[dim],
                                               alpha=self.b,
                                               method=self.int_transform,
                                               )
        
        return vec

    def evolute(self, ngen, x0=None, verbose=True):
        """
        This function evolutes the GWO algorithm for number of generations.
        
        :param ngen: (int) number of generations to evolute
        :param x0: (list of lists) initial position of the wolves (must be of same size as ``nwolves``)
        :param verbose: (bool) print statistics to screen
        
        :return: (dict) dictionary containing major GWO search results
        """
        self.history = {'alpha_wolf':[], 'beta_wolf':[], 'delta_wolf': [], 'fitness':[]}
        self.fitness_best=float("inf") 
        self.verbose=verbose
        self.Positions = np.zeros((self.nwolves, self.dim))
        if x0 is not None:
            assert len(x0) == self.nwolves, '--error: the length of x0 ({}) MUST equal the number of wolves in the group ({})'.format(len(x0), self.nwolves)
            for i in range(self.nwolves):
                self.Positions[i,:] = x0[i]
        else:
            #self.Positions=self.init_sample(self.bounds)  #TODO, update GWO laterfor mixed-integer optimisation
            # Initialize the positions of search agents
            
            for i in range(self.dim):
                self.Positions[:, i] = (np.random.uniform(0, 1, self.nwolves) * (self.ub[i] - self.lb[i]) + self.lb[i])       
 
        # initialize alpha, beta, and delta_pos
        Alpha_pos = np.zeros(self.dim)
        Alpha_score = float("inf")  #GWO is built to minimize
    
        Beta_pos = np.zeros(self.dim)
        Beta_score = float("inf")  #GWO is built to minimize
    
        Delta_pos = np.zeros(self.dim)
        Delta_score = float("inf") #GWO is built to minimize
                       
        for l in range(0, ngen):
            self.b= 1 - l * ((1) / ngen)  #mir: b decreases linearly between 1 to 0, for discrete mutation
            #---------------------
            # Fitness calcs
            #---------------------
            self.x_lst=[]
            for case in range (0, self.Positions.shape[0]):
                self.x_lst.append(list(self.Positions[case, :]))
        
            if self.ncores > 1:
                p=MyPool(self.ncores)
                self.fitness = p.map(self.fit, self.x_lst)
                p.close(); p.join()            
            else:
                self.fitness=[]
                for item in self.x_lst:
                    self.fitness.append(self.fit(item))  
                    
            #----------------------
            #  Update wolf scores
            #----------------------
            #Loop through the fitness list and update the score of alpha, beta, gamma, and omega!
            for i, fits in enumerate(self.fitness):
                # Update Alpha, Beta, and Delta
                if fits < Alpha_score:
                    Delta_score = Beta_score  # Update delte
                    Delta_pos = Beta_pos.copy()
                    Beta_score = Alpha_score  # Update beta
                    Beta_pos = Alpha_pos.copy()
                    Alpha_score = fits
                    # Update alpha
                    Alpha_pos = self.Positions[i, :].copy()
    
                if fits > Alpha_score and fits < Beta_score:
                    Delta_score = Beta_score  # Update delte
                    Delta_pos = Beta_pos.copy()
                    Beta_score = fits  # Update beta
                    Beta_pos = self.Positions[i, :].copy()
    
                if fits > Alpha_score and fits > Beta_score and fits < Delta_score:
                    Delta_score = fits  # Update delta
                    Delta_pos = self.Positions[i, :].copy()
                
                #save the best of the best!!!
                if fits < self.fitness_best:
                    self.fitness_best=fits
                    self.x_best=self.Positions[i, :].copy()
                
                
            self.history['alpha_wolf'].append(Alpha_score)
            self.history['beta_wolf'].append(Beta_score)
            self.history['delta_wolf'].append(Delta_score)
            
            a = 2 - l * ((2) / ngen)
            # a decreases linearly from 2 to 0
            
            #--------------------------------
            # Position update loop
            #--------------------------------
            # Update the position of search wolves
            for i in range(0, self.nwolves):
                for j in range(0, self.dim):
    
                    r1 = random.random()  # r1 is a random number in [0,1]
                    r2 = random.random()  # r2 is a random number in [0,1]
    
                    A1 = 2 * a * r1 - a
                    # Equation (3.3)
                    C1 = 2 * r2
                    # Equation (3.4)
    
                    D_alpha = abs(C1 * Alpha_pos[j] - self.Positions[i, j])
                    # Equation (3.5)-part 1
                    X1 = Alpha_pos[j] - A1 * D_alpha
                    # Equation (3.6)-part 1
    
                    r1 = random.random()
                    r2 = random.random()
    
                    A2 = 2 * a * r1 - a
                    # Equation (3.3)
                    C2 = 2 * r2
                    # Equation (3.4)
    
                    D_beta = abs(C2 * Beta_pos[j] - self.Positions[i, j])
                    # Equation (3.5)-part 2
                    X2 = Beta_pos[j] - A2 * D_beta
                    # Equation (3.6)-part 2
    
                    r1 = random.random()
                    r2 = random.random()
    
                    A3 = 2 * a * r1 - a
                    # Equation (3.3)
                    C3 = 2 * r2
                    # Equation (3.4)
    
                    D_delta = abs(C3 * Delta_pos[j] - self.Positions[i, j])
                    # Equation (3.5)-part 3
                    X3 = Delta_pos[j] - A3 * D_delta
                    # Equation (3.5)-part 3
    
                    self.Positions[i, j] = (X1 + X2 + X3) / 3  # Equation (3.7)
                
                self.Positions[i,:]=self.ensure_bounds(self.Positions[i,:])
                self.Positions[i, :] = self.ensure_discrete(self.Positions[i, :])

            #--mir
            if self.mode=='max':
                self.fitness_best_correct=-self.fitness_best
            else:
                self.fitness_best_correct=self.fitness_best
                
            # Print statistics
            if self.verbose and i % self.nwolves:
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
                print('GWO step {}/{}, nwolves={}, Ncores={}'.format((l+1)*self.nwolves, ngen*self.nwolves, self.nwolves, self.ncores))
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
                print('Best Group Fitness:', np.round(self.fitness_best_correct,6))
                print('Best Group Position:', np.round(self.x_best,6))
                print('Alpha wolf Fitness:', np.round(Alpha_score,6))
                print('Beta wolf Fitness:', np.round(Beta_score,6))
                print('Delta wolf Fitness:', np.round(Delta_score,6))
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')

    
            self.history['fitness'].append(Alpha_score)

        if self.verbose:
            print('------------------------ GWO Summary --------------------------')
            print('Best fitness (y) found:', self.fitness_best_correct)
            print('Best individual (x) found:', self.x_best)
            print('--------------------------------------------------------------') 
        
        
        #converted_fit=-np.array(self.fitness)  #this is to get back to original sign
        #-------------------------------------
        #return population ranked for PESA2
        #-------------------------------------
        pesa_pop=defaultdict(list)
        for i in range(len(self.x_lst)):
            pesa_pop[i].append(self.x_lst[i])
            if self.mode=='max':
                pesa_pop[i].append(-self.fitness[i])
            else:
                pesa_pop[i].append(self.fitness[i])
        
        #--mir
        if self.mode=='max':
            self.history['alpha_wolf']=[-item for item in self.history['alpha_wolf']]
            self.history['beta_wolf']=[-item for item in self.history['beta_wolf']]
            self.history['delta_wolf']=[-item for item in self.history['delta_wolf']]
            self.history['fitness']=[-item for item in self.history['fitness']]
            
        #print(pesa_pop)
        
        return self.x_best, self.fitness_best_correct, pesa_pop


# -*- coding: utf-8 -*-
"""
Created on Wed Feb 24 19:08:52 2021

@author: majdi
"""

import os
import sys
import subprocess
from setuptools import setup, find_packages
from distutils.version import LooseVersion
from neorl.version import version
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", message=r"Passing", category=FutureWarning)

#check python version
if sys.version_info.major != 3:
    raise ValueError('--ERROR: This package is only compatible with Python 3, but you are running '
          'Python {}. The installation will likely fail.'.format(sys.version_info.major))
    
# The directory containing this file
HERE = os.getcwd()

# The text of the README file
with open(os.path.join(HERE , "README.md"), encoding='utf-8') as f:
    README = f.read()
    
    
long_description = r"""
# NEORL

NEORL (**N**euro**E**volution **O**ptimization with **R**einforcement **L**earning) is a set of implementations of hybrid algorithms combining neural networks and evolutionary computation based on a wide range of machine learning and evolutionary intelligence architectures. NEORL aims to solve large-scale optimization problems relevant to operation & optimization research, engineering, business, and other disciplines. 

NEORL can be used for multidisciplinary applications for research, industrial, academic, and/or teaching purposes. NEORL can be used as a standalone platform or an additional benchmarking tool to supplement or validate other optimization packages. Our objective when we built NEORL is to give the user a simple and easy-to-use framework with an access to a wide range of algorithms, covering both standalone and hybrid algorithms in evolutionary, swarm, supervised learning, deep learning, and reinforcement learning. We hope NEORL will allow beginners to enjoy more advanced optimization and algorithms, without being involved in too many theoretical/implementation details, and give experts an opportunity to solve large-scale optimization problems.
## Copyright

This repository and its content are copyright of [Exelon Corporation](https://www.exeloncorp.com/) © in collaboration with [MIT](https://web.mit.edu/nse/) Nuclear Science and Engineering 2021. All rights reserved.

You can read the first successful application of NEORL for nuclear fuel optimisation in this [News Article](https://news.mit.edu/2020/want-cheaper-nuclear-energy-turn-design-process-game-1217).

## Links

Repository:
https://github.com/mradaideh/neorl

Main News Article:
https://news.mit.edu/2020/want-cheaper-nuclear-energy-turn-design-process-game-1217

Documentation:
https://neorl.readthedocs.io/en/latest/index.html

## Quick Example

Here is a quick example of how to use NEORL to minimize a 5-D sphere function:
```python
#---------------------------------
# Import packages
#---------------------------------
import numpy as np
import matplotlib.pyplot as plt
from neorl import DE, XNES

#---------------------------------
# Fitness
#---------------------------------
#Define the fitness function
def FIT(individual):
    '''Sphere test objective function.
            F(x) = sum_{i=1}^d xi^2
            d=1,2,3,...
            Range: [-100,100]
            Minima: 0
    '''

    return sum(x**2 for x in individual)

#---------------------------------
# Parameter Space
#---------------------------------
#Setup the parameter space (d=5)
nx=5
BOUNDS={}
for i in range(1,nx+1):
    BOUNDS['x'+str(i)]=['float', -100, 100]

#---------------------------------
# DE
#---------------------------------
de=DE(mode='min', bounds=BOUNDS, fit=FIT, npop=50, CR=0.5, F=0.7, ncores=1, seed=1)
x_best, y_best, de_hist=de.evolute(ngen=120, verbose=0)
print('---DE Results---', )
print('x:', x_best)
print('y:', y_best)

#---------------------------------
# NES
#---------------------------------
x0=[-50]*len(BOUNDS)
amat = np.eye(nx)
xnes=XNES(mode='min', bounds=BOUNDS, fit=FIT, npop=50, eta_mu=0.9,
          eta_sigma=0.5, adapt_sampling=True, seed=1)
x_best, y_best, nes_hist=xnes.evolute(120, x0=x0, verbose=0)
print('---XNES Results---', )
print('x:', x_best)
print('y:', y_best)


#---------------------------------
# Plot
#---------------------------------
#Plot fitness for both methods
plt.figure()
plt.plot(np.array(de_hist), label='DE')
plt.plot(np.array(nes_hist['fitness']), label='NES')
plt.xlabel('Generation')
plt.ylabel('Fitness')
plt.legend()
plt.show()
```

## Citing the Project

To cite this repository in publications:

```
@misc{neorl,
  author = {Radaideh, Majdi I. and Du, Katelin and Seurin, Paul and Seyler, Devin and Gu, Xubo and Wang, Haijia and Shirvan, Koroush},
  title = {NEORL},
  year = {2021},
  publisher = {GitHub},
  journal = {GitHub repository},
  howpublished = {\url{https://github.com/mradaideh/neorl}},
}
```


"""

# Read version from file
__version__ = version()
    
# Check tensorflow installation to avoid
# breaking pre-installed tf gpu ##(credit to @hill-a and stable-baselines)
def find_tf_dependency():
    install_tf, tf_gpu = False, False
    try:
        import tensorflow as tf
        if tf.__version__ < LooseVersion('1.8.0'):
            install_tf = True
            # check if a gpu version is needed
            tf_gpu = tf.test.is_gpu_available()
    except ImportError:
        install_tf = True
        # Check if a nvidia gpu is present
        for command in ['nvidia-smi', '/usr/bin/nvidia-smi', 'nvidia-smi.exe']:
            try:
                if subprocess.call([command]) == 0:
                    tf_gpu = True
                    break
            except IOError:  # command does not exist / is not executable
                pass
        if os.environ.get('USE_GPU') == 'True':  # force GPU even if not auto-detected
            tf_gpu = True

    tf_dependency = []
    if install_tf:
        tf_dependency = ['tensorflow-gpu>=1.8.0,<2.0.0'] if tf_gpu else ['tensorflow>=1.8.0,<2.0.0']
        if tf_gpu:
            print("A GPU was detected, tensorflow-gpu will be installed")

    return tf_dependency
    
# This call to setup() does all the work
setup(
    name="neorl",
    packages=[package for package in find_packages() if package.startswith('neorl')],
    include_package_data=True,
    package_data={'neorl': ['requirements.txt', 'version.txt']},
    install_requires=['tensorflow==1.14.0',
                      'numpy== 1.16.2',
                      'gym >= 0.15.4, < 0.17.0',
                      'scikit-optimize==0.8.1',
                      'cloudpickle >= 1.2.2',
                      'h5py < 2.10.0',    #for windows version (NHHO fails with model load)
					  'scikit-learn <= 0.24.2',
                      'neat-python', 
                      'xarray==0.16',
                      'protobuf==3.20',
                      'scipy',
                      'joblib',
                      'pandas',
                      'xlrd==1.2.0',
                      'matplotlib',
                      'pytest',
                      'pytest-cov',
                      'sphinx',
                      'sphinx-rtd-theme',
                      'sphinx-autobuild'] + find_tf_dependency(),               
     extras_require={'tests': ['pytest', 'pytest-cov', 'pytest-env', 'pytest-xdist', 'pytype'],
                     'docs': ['sphinx', 'sphinx-autobuild', 'sphinx-rtd-theme']},   
    
    description="NeuroEvolution Optimisation with Reinforcement Learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mradaideh/neorl",
    author="Majdi I. Radaideh",
    author_email="radaideh@mit.edu",
    entry_points={
        "console_scripts": [
            "neorl=neorl.scripts:main ",
        ]
    },
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"],
    version= __version__,
)



import numpy as np
import neorl.benchmarks.cec17 as functions    #import all cec17 functions
from neorl import DE

def test_benchmarks2():
    
    reduced_func=functions.all_functions[:10] #keep only the first 10 functions
    nx = 2 #set dimension
    BOUNDS={}
    for i in range(1,nx+1):
        BOUNDS['x'+str(i)]=['float', -100, 100]
    
    for FIT in reduced_func:
        #setup and evolute PSO
        de=DE(mode='min', bounds=BOUNDS, fit=FIT, npop=60, F=0.5, 
              CR=0.7, ncores=1, seed=1)
        x_best, y_best, de_hist=de.evolute(ngen=100, verbose=0)
        opt=float(FIT.__name__.strip('f'))*100
        print('Function: {}, x-DE={}, y-DE={}, y-Optimal={}'.format(FIT.__name__, 
                                                                 np.round(x_best,2), 
                                                                 np.round(y_best,2), 
                                                                 opt))
        
test_benchmarks2()


import numpy as np


class Dataset(object):
    def __init__(self, data_map, shuffle=True):
        """
        Data loader that handles batches and shuffling.
        WARNING: this will alter the given data_map ordering, as dicts are mutable

        :param data_map: (dict) the input data, where every column is a key
        :param shuffle: (bool) Whether to shuffle or not the dataset
            Important: this should be disabled for recurrent policies
        """
        self.data_map = data_map
        self.shuffle = shuffle
        self.n_samples = next(iter(data_map.values())).shape[0]
        self._next_id = 0
        if self.shuffle:
            self.shuffle_dataset()

    def shuffle_dataset(self):
        """
        Shuffles the data_map
        """
        perm = np.arange(self.n_samples)
        np.random.shuffle(perm)

        for key in self.data_map:
            self.data_map[key] = self.data_map[key][perm]

    def next_batch(self, batch_size):
        """
        returns a batch of data of a given size

        :param batch_size: (int) the size of the batch
        :return: (dict) a batch of the input data of size 'batch_size'
        """
        if self._next_id >= self.n_samples:
            self._next_id = 0
            if self.shuffle:
                self.shuffle_dataset()

        cur_id = self._next_id
        cur_batch_size = min(batch_size, self.n_samples - self._next_id)
        self._next_id += cur_batch_size

        data_map = dict()
        for key in self.data_map:
            data_map[key] = self.data_map[key][cur_id:cur_id + cur_batch_size]
        return data_map

    def iterate_once(self, batch_size):
        """
        generator that iterates over the dataset

        :param batch_size: (int) the size of the batch
        :return: (dict) a batch of the input data of size 'batch_size'
        """
        if self.shuffle:
            self.shuffle_dataset()

        while self._next_id <= self.n_samples - batch_size:
            yield self.next_batch(batch_size)
        self._next_id = 0

    def subset(self, num_elements, shuffle=True):
        """
        Return a subset of the current dataset

        :param num_elements: (int) the number of element you wish to have in the subset
        :param shuffle: (bool) Whether to shuffle or not the dataset
        :return: (Dataset) a new subset of the current Dataset object
        """
        data_map = dict()
        for key in self.data_map:
            data_map[key] = self.data_map[key][:num_elements]
        return Dataset(data_map, shuffle)


def iterbatches(arrays, *, num_batches=None, batch_size=None, shuffle=True, include_final_partial_batch=True):
    """
    Iterates over arrays in batches, must provide either num_batches or batch_size, the other must be None.

    :param arrays: (tuple) a tuple of arrays
    :param num_batches: (int) the number of batches, must be None is batch_size is defined
    :param batch_size: (int) the size of the batch, must be None is num_batches is defined
    :param shuffle: (bool) enable auto shuffle
    :param include_final_partial_batch: (bool) add the last batch if not the same size as the batch_size
    :return: (tuples) a tuple of a batch of the arrays
    """
    assert (num_batches is None) != (batch_size is None), 'Provide num_batches or batch_size, but not both'
    arrays = tuple(map(np.asarray, arrays))
    n_samples = arrays[0].shape[0]
    assert all(a.shape[0] == n_samples for a in arrays[1:])
    inds = np.arange(n_samples)
    if shuffle:
        np.random.shuffle(inds)
    sections = np.arange(0, n_samples, batch_size)[1:] if num_batches is None else num_batches
    for batch_inds in np.array_split(inds, sections):
        if include_final_partial_batch or len(batch_inds) == batch_size:
            yield tuple(a[batch_inds] for a in arrays)



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Tue Feb 25 14:42:24 2020
#@author: Majdi Radaideh
#"""

import warnings
warnings.filterwarnings("ignore")
import random
#import gym
import pandas as pd
import numpy as np
#from deap import algorithms, base, creator, tools
# import input parameters from the user
#from neorl.parsers.PARSER import InputChecker
from neorl.evolu.crossover import cxES2point, mutES, select
from collections import defaultdict

import multiprocessing
import multiprocessing.pool
from neorl.utils.seeding import set_neorl_seed

#import os, csv
import copy

class NoDaemonProcess(multiprocessing.Process):
    # make 'daemon' attribute always return False
    def _get_daemon(self):
        return False
    def _set_daemon(self, value):
        pass
    daemon = property(_get_daemon, _set_daemon)

# We sub-class multiprocessing.pool.Pool instead of multiprocessing.Pool
# because the latter is only a wrapper function, not a proper class.
class MyPool(multiprocessing.pool.Pool):
    Process = NoDaemonProcess

class GA:
    """
    Parallel Genetic Algorithms
    
    :param bounds: (dict) input parameter type and lower/upper bounds in dictionary form. Example: {'x1': ['int', 1, 4], 'x2': ['float', 0.1, 0.8], 'x3': ['float', 2.2, 6.2]}
    :param fit: (function) the fitness function 
    :param npop: (int) number of individuals in the population
    :param cxpb: (float) population crossover probability between [0,1]
    :param mutpb: (float) population mutation probability between [0,1]
    :param chi: (float) individual mutation probability between [0,1]
    :param ncores: (int) number of parallel processors
    :param seed: (int) random seed for sampling
    """
    def __init__ (self, bounds, fit, npop=50, mu=50, cxpb=0.7, cxfunc='cx2point', 
                  mutfunc='', mutpb=0.2, 
                  chi=0.1, ncores=1, seed=None):    

        
        set_neorl_seed(seed)
        
        self.bounds=bounds
        self.fit=fit
        self.npop=npop
        self.mu=mu
        self.cxpb=cxpb
        self.mutpb=mutpb
        self.chi=chi
        self.ncores=ncores
        self.seed=seed
        
        assert mu <= npop, '--error: the value of `mu` must be less than or equal `npop`'
        assert 0 <= chi <= 1, '--error: `chi` must be between [0,1]'
        assert 0 <= cxpb <= 1, '--error: `cxpb` must be between [0,1]'
        assert 0 <= mutpb <= 1, '--error: `mutpb` must be between [0,1]'
            
    def gen_object(self, inp):
        #"""
        #This is a worker for multiprocess Pool
        #Inputs:
        #    inp (list of lists): contains data for each core [[ind1, caseid1], ...,  [indN, caseidN]]
        #Returns:
        #    fitness value (float)
        #"""
        return self.env.fit(inp)

    def GenES(self, lb, datatype, smin, smax, ub=None):
        #"""
        #Individual generator
        #Inputs:
        #    -lb,ub,datatype (list): input paramter lower/upper bounds in dictionary form
        #Returns 
        #    -ind (list): an individual vector with values sampled from bounds
        #    -strategy (list): the strategy vector with values between smin and smax
        #"""
        size = len(lb) # size of individual 
        content=[]
        
        if self.kbs_path:
            assert self.kbs_data.shape[0] > self.lambda_, '--error: the size of the KBS dataset ({}) is not enough to cover initial population lambda ({})'.format(self.kbs_data.shape[0],self.lambda_)
            index=random.randint(0,self.kbs_data.shape[0]-1)
            ind=list(self.kbs_data[index,:])
        else:
            for i in range (len(lb)):
                if datatype[i] == 'int':
                    content.append(random.randint(lb[i], ub[i]))
                elif datatype[i] == 'float':
                    content.append(random.uniform(lb[i], ub[i]))
                elif datatype[i] == 'grid':
                    content.append(random.sample(lb[i],1)[0])
                else:
                    raise Exception ('unknown data type is given, either int, float, or grid are allowed for parameter bounds')
            
            ind=list(content)
        strategy = [random.uniform(smin,smax) for _ in range(size)]
        return ind, strategy

    def init_pop(self, warmup):
        #"""
        #Population intializer 
        #Inputs:
        #    -warmup (int): number of individuals to create and evaluate initially
        #Returns 
        #    -pop (dict): initial population in a dictionary form, looks like this
            
        #"""
        #initialize the population and strategy and run them in parallel (these samples will be used to initialize the memory)
        pop=defaultdict(list)
        # dict key runs from 0 to warmup-1
        # index 0: individual, index 1: strategy, index 2: fitness 
        # Example: 
        """
        pop={key: [ind, strategy, fitness]}
        pop={0: [[1,2,3,4,5], [0.1,0.2,0.3,0.4,0.5], 1.2], 
             ... 
             99: [[1.1,2.1,3.1,4.1,5.1], [0.1,0.2,0.3,0.4,0.5], 5.2]}
        """
        for i in range (warmup):
            #caseid='es_gen{}_ind{}'.format(0,i+1)  #caseid are only for logging purposes to distinguish sample source
            data=self.GenES(lb=self.lbound, ub=self.ubound, datatype=self.datatype, smax=self.smax, smin=self.smin)
            pop[i].append(data[0])
            pop[i].append(data[1])
        
        if self.ncores > 1:  #evaluate warmup in parallel
            core_list=[]
            for key in pop:
                core_list.append(pop[key][0])
            p=MyPool(self.ncores)
            fitness = p.map(self.gen_object, core_list)
            p.close()
            p.join()
            
            [pop[ind].append(fitness[ind]) for ind in range(len(pop))]
        
        else: #evaluate warmup in series
            for key in pop:
                #caseid='es_gen{}_ind{}'.format(0,key+1)
                fitness=self.env.fit(pop[key][0])
                pop[key].append(fitness)
        
        return pop  #return final pop dictionary with ind, strategy, and fitness

    def select(pop, k=1):
        """
        Select function sorts the population from max to min based on fitness and select k best
        Inputs:
            pop (dict): population in dictionary structure
            k (int): top k individuals are selected
        Returns:
            best_dict (dict): the new ordered dictionary with top k selected 
        """
        
        pop=list(pop.items())
        pop.sort(key=lambda e: e[1][2], reverse=True)
        sorted_dict=dict(pop[:k])
        
        #This block creates a new dict where keys are reset to 0 ... k in order to avoid unordered keys after sort
        best_dict=defaultdict(list)
        index=0
        for key in sorted_dict:
            best_dict[index].append(sorted_dict[key][0])
            best_dict[index].append(sorted_dict[key][1])
            best_dict[index].append(sorted_dict[key][2])
            index+=1
        
        sorted_dict.clear()
        return best_dict

    def GenOffspring(self, pop):
        #"""
        #This function generates the offspring by applying crossover, mutation **or** reproduction. 
        #The sum of both probabilities self.cxpb and self.mutpb must be in [0,1]
        #The reproduction probability is 1 - cxpb - mutpb
        #The new offspring goes for fitness evaluation
        #Inputs:
        #    pop (dict): population in dictionary structure
        #Returns:
        #    offspring (dict): new modified population in dictionary structure    
        #"""
        pop_indices=list(range(0,len(pop)))
        offspring = defaultdict(list)
        for i in range(self.lambda_):
            alpha = random.random()
            #------------------------------
            # Crossover
            #------------------------------
            if alpha < self.cxpb:            
                index1, index2 = random.sample(pop_indices,2)
                ind1, ind2, strat1, strat2 = cx2point(ind1=list(pop[index1][0]),ind2=list(pop[index2][0]), 
                                                     strat1=list(pop[index1][1]),strat2=list(pop[index2][1]))
                offspring[i].append(ind1)
                offspring[i].append(strat1)
                #print('crossover is done for sample {} between {} and {}'.format(i,index1,index2))
            #------------------------------
            # Mutation
            #------------------------------
            elif alpha < (self.cxpb + self.mutpb):  # Apply mutation
                index = random.choice(pop_indices)
                ind, strat=mutES(ind=list(pop[index][0]), strat=list(pop[index][1]), lb=self.lbound, ub=self.ubound, datatype=self.datatype, smax=self.smax, smin=self.smin)
                offspring[i].append(ind)
                offspring[i].append(strat)
                #print('mutation is done for sample {} based on {}'.format(i,index))
            #------------------------------
            # Reproduction from population
            #------------------------------
            else:                         
                index=random.choice(pop_indices)
                offspring[i].append(pop[index][0])
                offspring[i].append(pop[index][1])
                #print('reproduction is done for sample {} based on {}'.format(i,index))
    
        return offspring

    def evolute(self, ngen, x0, verbose=0):
        """
        This function evolutes the GA algorithm for number of generations.
        
        :param ngen: (int) number of generations to evolute
        :param x0: (list of lists) the initial individuals of the population
        :param verbose: (bool) print statistics to screen
        
        :return: (dict) dictionary containing major GA search results
        """
        set_neorl_seed(self.seed)

        if x0:
            assert len(x0) == self.npop, '--error: the length of `x0` ({}) (initial population) must equal to number of individuals `npop` ({})'.format(len(x0), self.npop)
            population = self.InitPopulation(x0=x0)
        else:
            population = self.InitPopulation()
        
        # Begin the evolution process
        for gen in range(1, ngen + 1):
            
            # Vary the population and generate new offspring
            offspring = self.GenOffspring(pop=population)
            
            # Evaluate the individuals with an invalid fitness with multiprocessign Pool
            # create and run the Pool
            if self.ncores > 1:
                core_list=[]
                for key in offspring:
                    core_list.append(offspring[key][0])
                #initialize a pool
                print('---------- Start the real generations------')
                p=MyPool(self.ncores)
                fitness = p.map(self.gen_object, core_list)
                p.close(); p.join()
                
                [offspring[ind].append(fitness[ind]) for ind in range(len(offspring))]
            else: #serial calcs
                for ind in range(len(offspring)):
                    fitness=self.env.fit(offspring[ind][0])
                    offspring[ind].append(fitness)
                                
            # Select the next generation population
            #print([(offspring[item][0],offspring[item][2]) for item in offspring])
            #print('-------------------------------')
            population = copy.deepcopy(select(pop=offspring, k=self.mu))
            #print([(population[item][0],population[item][2]) for item in population])
            #print('-------------------------------')        
            #print([(population[item][0],population[item][2]) for item in population])
            #print('-------------------------------')
                            
            #------------
            # Monitor Progress
            #------------
            if  gen % self.check_freq == 0 or gen == self.ngen:

                inds, rwd=[population[i][0] for i in population], [population[i][2] for i in population]
                mean_strategy=[np.mean(population[i][1]) for i in population]
                #------------
                # plot progress 
                #------------
                self.callback.plot_progress('Generation')
                
                if verbose:
                    print('############################################################')
                    print('ES generation {}/{}, CX={}, MUT={}, MU={}, LAMBDA={}'.format(gen,ngen, np.round(self.cxpb,2), np.round(self.mutpb,2), self.mu, self.lambda_))
                    print('############################################################')
                    print('Statistics for generation {}'.format(gen))
                    print('Best Reward:', np.round(np.max(rwd),2))
                    print('Mean Reward:', np.round(np.mean(rwd),2))
                    print('Max Strategy:', np.round(np.max(mean_strategy),3))
                    print('Min Strategy:', np.round(np.min(mean_strategy),3))
                    print('Average Strategy:', np.round(np.mean(mean_strategy),3))
                    print('############################################################')

            #------------
            # Update the population with KBS 
            #------------
            #print([(population[item][0],population[item][2]) for item in population])            
            if self.kbs_path:
                kbs_pop_indices=random.sample(range(self.kbs_data.shape[0]),self.kbs_pop)
                last_index=list(population.keys())
                last_index=last_index[-1]
                for i in range (len(kbs_pop_indices)):
                    ind_vec=list(self.kbs_data[kbs_pop_indices[i],:])
                    
                    if self.kbs_append: #append KBS individuals to the population
                        strategy = [random.uniform(self.smin,self.smax) for _ in range(len(self.lbound))]
                        kbs_ind=[ind_vec,strategy,0]
                        population[last_index+(i+1)]=kbs_ind
                    else: #replace the worst individuals in pop with KBS individuals
                        #population[last_index-i-1]=kbs_ind
                        #print(last_index-i)
                        population[last_index-i][0]=ind_vec
                        #print(population[last_index-i][0])
                    
        return population
        
#    def build(self):
#        
#        """
#        This function builds a GA module based on the passed user parameters
#        """
#        
#        random.seed(42)
#        pop0=self.init_pop(warmup=self.lambda_)
#        self.evolute(population=pop0, ngen=self.ngen, verbose=0)


from neorl import ACO
import random

def test_aco():
    
    #Define the fitness function
    def FIT(individual):
        """Sphere test objective function.
                        F(x) = sum_{i=1}^d xi^2
                        d=1,2,3,...
                        Range: [-100,100]
                        Minima: 0
        """
        y=sum(x**2 for x in individual)
        return y
    
    #Setup the parameter space (d=5)
    nx=5
    BOUNDS={}
    for i in range(1,nx+1):
        BOUNDS['x'+str(i)]=['float', -100, 100]
    
    nants=40
    x0=[[random.uniform(-100,100)]*nx for item in range(nants)]
    acor = ACO(mode='min', fit=FIT, bounds=BOUNDS, nants=nants, narchive=10, 
               Q=0.5, Z=1, ncores=1, seed=1)
    x_best, y_best, acor_hist=acor.evolute(ngen=100, x0=x0, verbose=1)

test_aco()


import numpy as np
from pathlib import Path
import pandas as pd
from tensorflow.keras.models import load_model
import sys
import inspect

cpath = Path(inspect.getfile(sys.modules[__name__])).resolve().parent

def transform_features(x, f="cos"):
    if f == "cos":
        return np.cos(x)
    elif f == "sin":
        return np.sin(x)
    elif f == "tanh":
        return np.tanh(x)

class QPowerModel:
    """
    Use to evaluate quadrant power splits from control drum configurations.
    Set up as init, then separately use method call to minimize reading times.
    """
    def __init__(self):
        #Find and load file
        model_file = cpath / Path("tools/microreactor_power_model.h5")
        self.raw_model = load_model(model_file)

    def eval(self, pert):
        pert2 = pert.copy()
        pertn = np.array([pert2, ])
        unorm = self.raw_model.predict(pertn).flatten()
        return unorm/unorm.sum()

def qPowerModel(pert):
    """Wrapper for QPowerModel that initializes and runs"""
    a = QPowerModel()
    return a.eval(pert)

if __name__ == "__main__":
    thetas = np.zeros(8)
    thetas[[6,7]] -= np.pi
    print(qPowerModel(thetas))



from neorl.tune import GRIDTUNE
from neorl import ES

def test_grid():
    
    #Define the fitness function (for original optimisation)
    def sphere(individual):
        y=sum(x**2 for x in individual)
        return y 
    
    def tune_fit(cxpb, mutpb, alpha):
    
        #--setup the parameter space
        nx=5
        BOUNDS={}
        for i in range(1,nx+1):
            BOUNDS['x'+str(i)]=['float', -100, 100]
    
        #--setup the ES algorithm
        es=ES(mode='min', bounds=BOUNDS, fit=sphere, lambda_=80, mu=40, mutpb=mutpb, alpha=alpha,
             cxmode='blend', cxpb=cxpb, ncores=1, seed=1)
    
        #--Evolute the ES object and obtains y_best
        #--turn off verbose for less algorithm print-out when tuning
        x_best, y_best, es_hist=es.evolute(ngen=100, verbose=0)
    
        return y_best #returns the best score
    
    param_grid={
    #def tune_fit(cxpb, mutpb, alpha):
    'cxpb': [0.2, 0.4],  #cxpb is first
    'mutpb': [0.05, 0.1],  #mutpb is second
    'alpha': [0.1, 0.2, 0.3, 0.4]}  #alpha is third
    
    gtune=GRIDTUNE(param_grid=param_grid, fit=tune_fit)
    gridres=gtune.tune(ncores=1)
    print(gridres)

test_grid()